import itertools
import random
import string
import warnings

import numpy as np
import pandas as pd

from plotnine.data import mtcars
from plotnine.utils import (
    _margins,
    add_margins,
    join_keys,
    match,
    ninteraction,
    pivot_apply,
    remove_missing,
    uniquecols,
)


def test__margins():
    vars = [("vs", "am"), ("gear",)]
    lst = _margins(vars, True)
    assert lst == [
        [],
        ["vs", "am"],
        ["am"],
        ["gear"],
        ["vs", "am", "gear"],
        ["am", "gear"],
    ]

    lst = _margins(vars, False)
    assert lst == []

    lst = _margins(vars, ["vs"])
    assert lst == [[], ["vs", "am"]]

    lst = _margins(vars, ["am"])
    assert lst == [[], ["am"]]

    lst = _margins(vars, ["vs", "am"])
    assert lst == [[], ["vs", "am"], ["am"]]

    lst = _margins(vars, ["gear"])
    assert lst == [[], ["gear"]]


def test_add_margins():
    df = mtcars.loc[:, ["mpg", "disp", "vs", "am", "gear"]]
    n = len(df)
    all_lst = ["(all)"] * n

    vars = [("vs", "am"), ("gear",)]
    dfx = add_margins(df, vars, True)

    assert dfx["vs"].dtype == "category"
    assert dfx["am"].dtype == "category"
    assert dfx["gear"].dtype == "category"

    # What we expect, where each row is of
    # column length n
    #
    # mpg   disp   vs     am     gear
    # ---   ----   --     --     ----
    # *     *      *      *      *
    # *     *      (all)  (all)  *
    # *     *      *      (all)  *
    # *     *      *      *      (all)
    # *     *      (all)  (all)  (all)
    # *     *      *      (all)  (all)

    assert all(dfx.loc[0 : n - 1, "am"] != all_lst)
    assert all(dfx.loc[0 : n - 1, "vs"] != all_lst)
    assert all(dfx.loc[0 : n - 1, "gear"] != all_lst)

    assert all(dfx.loc[n : 2 * n - 1, "vs"] == all_lst)
    assert all(dfx.loc[n : 2 * n - 1, "am"] == all_lst)

    assert all(dfx.loc[2 * n : 3 * n - 1, "am"] == all_lst)

    assert all(dfx.loc[3 * n : 4 * n - 1, "gear"] == all_lst)

    assert all(dfx.loc[4 * n : 5 * n - 1, "am"] == all_lst)
    assert all(dfx.loc[4 * n : 5 * n - 1, "vs"] == all_lst)
    assert all(dfx.loc[4 * n : 5 * n - 1, "gear"] == all_lst)

    assert all(dfx.loc[5 * n : 6 * n - 1, "am"] == all_lst)
    assert all(dfx.loc[5 * n : 6 * n - 1, "gear"] == all_lst)


def test_ninteraction():
    simple_vectors = [
        list(string.ascii_lowercase),
        random.sample(string.ascii_lowercase, 26),
        list(range(1, 27)),
    ]

    # vector of unique values is equivalent to rank
    for case in simple_vectors:
        df = pd.DataFrame(case)
        rank = df.rank(method="min")
        rank = rank[0].astype(int).tolist()
        rank_df = ninteraction(df)
        assert rank == rank_df

    # duplicates are numbered sequentially
    # df                    ids
    # [6, 6, 4, 4, 5, 5] -> [3, 3, 1, 1, 2, 2]
    for case in simple_vectors:
        rank = pd.DataFrame(case).rank(method="min")
        rank = rank[0].astype(int).repeat(2).tolist()
        rank_df = ninteraction(pd.DataFrame(np.array(case).repeat(2)))
        assert rank == rank_df

    # grids are correctly ranked
    df = pd.DataFrame(list(itertools.product([1, 2], range(1, 11))))
    assert ninteraction(df) == list(range(1, len(df) + 1))
    assert ninteraction(df, drop=True) == list(range(1, len(df) + 1))

    # zero length dataframe
    df = pd.DataFrame()
    assert ninteraction(df) == []

    # dataframe with single variable
    df = pd.DataFrame({"a": ["a"]})
    assert ninteraction(df) == [1]

    df = pd.DataFrame({"a": ["b"]})
    assert ninteraction(df) == [1]


def test_ninteraction_datetime_series():
    # When a pandas datetime is converted Numpy datetime, the two
    # no longer compare as equal! This test ensures that case is
    # not happening
    lst = ["2020-01-01", "2020-01-02", "2020-01-03"]
    df1 = pd.DataFrame(
        {
            "x": list("abcaabbcc"),
            "date_list": lst * 3,
        }
    )
    df2 = pd.DataFrame(
        {
            "x": list("abcaabbcc"),
            "date_list": pd.to_datetime(lst * 3),
        }
    )

    assert ninteraction(df1) == ninteraction(df2)


def test_join_keys():
    df1 = pd.DataFrame(
        {
            "a": [0, 0, 1, 1, 2, 2],
            "b": [0, 1, 2, 3, 1, 2],
            "c": [0, 1, 2, 3, 4, 5],
        }
    )

    # same array and columns the keys should be the same
    keys = join_keys(df1, df1, ["a", "b"])
    assert list(keys["x"]) == [1, 2, 3, 4, 5, 6]
    assert list(keys["x"]) == [1, 2, 3, 4, 5, 6]

    # Every other element of df2['b'] is changed
    # so every other key should be different
    df2 = pd.DataFrame(
        {
            "a": [0, 0, 1, 1, 2, 2],
            "b": [0, 11, 2, 33, 1, 22],
            "c": [1, 2, 3, 4, 5, 6],
        }
    )

    keys = join_keys(df1, df2, ["a", "b"])
    assert list(keys["x"]) == [1, 2, 4, 5, 7, 8]
    assert list(keys["y"]) == [1, 3, 4, 6, 7, 9]


def test_match():
    v1 = [1, 1, 2, 2, 3, 3]
    v2 = [1, 2, 3]
    v3 = [0, 1, 2]
    c = [1]

    assert list(match(v1, v2)) == [0, 0, 1, 1, 2, 2]
    assert list(match(v1, v2, incomparables=c)) == [-1, -1, 1, 1, 2, 2]
    assert list(match(v1, v3)) == [1, 1, 2, 2, -1, -1]


def test_uniquecols():
    df = pd.DataFrame(
        {
            "x": [1, 2, 3, 4],
            "y": ["a", "b", "c", "d"],
            "z": [8] * 4,
            "other": ["same"] * 4,
        }
    )
    df2 = pd.DataFrame({"z": [8], "other": ["same"]})
    result = uniquecols(df)
    assert result.equals(df2)


def test_remove_missing():
    df = pd.DataFrame({"a": [1.0, np.NaN, 3, np.inf], "b": [1, 2, 3, 4]})
    df2 = pd.DataFrame({"a": [1.0, 3, np.inf], "b": [1, 3, 4]})
    df3 = pd.DataFrame({"a": [1.0, 3], "b": [1, 3]})

    with warnings.catch_warnings(record=True) as w:
        res = remove_missing(df, na_rm=True, vars=["b"])
        res.equals(df)

        res = remove_missing(df)
        res.equals(df2)

        res = remove_missing(df, na_rm=True, finite=True)
        res.equals(df3)
        assert len(w) == 1


def test_pivot_apply():
    df = pd.DataFrame(
        {
            "id": list("abcabc"),
            "x": [1, 2, 3, 11, 22, 33],
            "y": [1, 2, 3, 11, 22, 33],
        }
    )

    res1 = pivot_apply(df, "x", "id", np.min)
    res2 = pivot_apply(df, "y", "id", np.max)

    assert res1.index.tolist() == list("abc")
    assert res1.index.name == "id"
    assert (res1 + res2 == [12, 24, 36]).all()



from ..doctools import document
from .geom_point import geom_point


@document
class geom_sina(geom_point):
    """
    Draw a sina plot

    A sina plot is a data visualization chart suitable for plotting
    any single variable in a multiclass dataset. It is an enhanced
    jitter strip chart, where the width of the jitter is controlled
    by the density distribution of the data within each class.

    {usage}

    Parameters
    ----------
    {common_parameters}

    See Also
    --------
    plotnine.stats.stat_sina

    References
    ----------
    Sidiropoulos, N., S. H. Sohi, T. L. Pedersen, B. T. Porse, O. Winther,
    N. Rapin, and F. O. Bagger. 2018.
    "SinaPlot: An Enhanced Chart for Simple and Truthful Representation of
    Single Observations over Multiple Classes."
    J. Comp. Graph. Stat 27: 673–76.
    """

    DEFAULT_PARAMS = {"stat": "sina", "position": "dodge", "na_rm": False}



from __future__ import annotations

import typing

from ..doctools import document
from .geom import geom

if typing.TYPE_CHECKING:
    from typing import Any

    import pandas as pd

    from plotnine.iapi import panel_view
    from plotnine.typing import Axes, Coord


@document
class geom_blank(geom):
    """
    An empty plot

    {usage}

    Parameters
    ----------
    {common_parameters}
    """

    DEFAULT_PARAMS = {
        "stat": "identity",
        "position": "identity",
        "na_rm": False,
    }

    def draw_panel(
        self,
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        pass

    def handle_na(self, data: pd.DataFrame) -> pd.DataFrame:
        return data



from __future__ import annotations

import typing

import pandas as pd

from ..exceptions import PlotnineError
from ..geoms.geom import geom as geom_base_class
from ..mapping import aes
from ..mapping.aes import POSITION_AESTHETICS
from ..utils import Registry, is_scalar_or_string

if typing.TYPE_CHECKING:
    from typing import Any

    from plotnine.typing import Ggplot, Layer


class annotate:
    """
    Create an annotation layer

    Parameters
    ----------
    geom : geom or str
        geom to use for annotation, or name of geom (e.g. 'point').
    x : float
        Position
    y : float
        Position
    xmin : float
        Position
    ymin : float
        Position
    xmax : float
        Position
    ymax : float
        Position
    xend : float
        Position
    yend : float
        Position
    xintercept : float
        Position
    yintercept : float
        Position
    kwargs : dict
        Other aesthetics or parameters to the geom.

    Notes
    -----
    The positioning aethetics ``x, y, xmin, ymin, xmax, ymax, xend, yend,
    xintercept, yintercept`` depend on which `geom` is used.

    You should choose or ignore accordingly.

    All `geoms` are created with :code:`stat='identity'`.
    """

    _annotation_geom: geom_base_class

    def __init__(
        self,
        geom: str | type[geom_base_class],
        x: float | None = None,
        y: float | None = None,
        xmin: float | None = None,
        xmax: float | None = None,
        xend: float | None = None,
        xintercept: float | None = None,
        ymin: float | None = None,
        ymax: float | None = None,
        yend: float | None = None,
        yintercept: float | None = None,
        **kwargs: Any,
    ):
        variables = locals()

        # position only, and combined aesthetics
        pos_aesthetics = {
            loc: variables[loc]
            for loc in POSITION_AESTHETICS
            if variables[loc] is not None
        }
        aesthetics = pos_aesthetics.copy()
        aesthetics.update(kwargs)

        # Check if the aesthetics are of compatible lengths
        lengths, info_tokens = [], []
        for ae, val in aesthetics.items():
            if is_scalar_or_string(val):
                continue
            lengths.append(len(val))
            info_tokens.append((ae, len(val)))

        if len(set(lengths)) > 1:
            details = ", ".join([f"{n} ({l})" for n, l in info_tokens])
            msg = f"Unequal parameter lengths: {details}"
            raise PlotnineError(msg)

        # Stop pandas from complaining about all scalars
        if all(is_scalar_or_string(val) for val in pos_aesthetics.values()):
            for ae in pos_aesthetics.keys():
                pos_aesthetics[ae] = [pos_aesthetics[ae]]
                break

        data = pd.DataFrame(pos_aesthetics)
        if isinstance(geom, str):
            geom_klass: type[geom_base_class] = Registry[f"geom_{geom}"]
        elif isinstance(geom, type) and issubclass(geom, geom_base_class):
            geom_klass = geom
        else:
            raise PlotnineError(
                "geom must either be a plotnine.geom.geom() "
                "descendant (e.g. plotnine.geom_point), or "
                "a string naming a geom (e.g. 'point', 'text', "
                f"...). Got {repr(geom)}"
            )

        mappings = aes(**{str(ae): ae for ae in data.columns})

        # The positions are mapped, the rest are manual settings
        self._annotation_geom = geom_klass(
            mappings,
            data,
            stat="identity",
            inherit_aes=False,
            show_legend=False,
            **kwargs,
        )

    def __radd__(self, gg: Ggplot) -> Ggplot:
        """
        Add to ggplot
        """
        gg += self.to_layer()  # Add layer
        return gg

    def to_layer(self) -> Layer:
        """
        Make a layer that represents this annotation

        Returns
        -------
        out : layer
            Layer
        """
        return self._annotation_geom.to_layer()



from __future__ import annotations

import typing

from mizani.scale import scale_continuous, scale_discrete

if typing.TYPE_CHECKING:
    from typing import Any, Sequence

    from plotnine.typing import TupleFloat2


class Range:
    """
    Base class for all ranges
    """

    #: Holds the range information after training
    range: TupleFloat2

    def reset(self):
        """
        Reset range
        """
        del self.range

    def train(self, x: Sequence[Any]):
        """
        Train range
        """
        raise NotImplementedError("Not Implemented.")

    def is_empty(self) -> bool:
        """
        Whether there is range information
        """
        return not hasattr(self, "range")


class RangeContinuous(Range):
    """
    Continuous Range
    """

    range: TupleFloat2

    def train(self, x: Sequence[Any]):
        """
        Train continuous range
        """
        rng = None if self.is_empty() else self.range
        self.range = scale_continuous.train(x, rng)  # pyright: ignore


class RangeDiscrete(Range):
    """
    Discrete Range
    """

    range: Sequence[Any]

    def train(self, x: Sequence[Any], drop: bool = False, na_rm: bool = False):
        """
        Train discrete range
        """
        rng = None if self.is_empty() else self.range
        self.range = scale_discrete.train(
            x, rng, drop, na_rm=na_rm
        )  # pyright: ignore



import os

import pytest

from plotnine import (
    aes,
    element_blank,
    element_line,
    element_text,
    facet_grid,
    facet_wrap,
    geom_point,
    ggplot,
    labs,
    theme,
    theme_538,
    theme_bw,
    theme_classic,
    theme_dark,
    theme_gray,
    theme_light,
    theme_linedraw,
    theme_matplotlib,
    theme_minimal,
    theme_seaborn,
    theme_tufte,
    theme_void,
    theme_xkcd,
)
from plotnine.data import mtcars


def test_add_complete_complete():
    theme1 = theme_gray()
    theme2 = theme_matplotlib()
    theme3 = theme1 + theme2
    assert theme3 == theme2


def test_add_complete_partial():
    theme1 = theme_gray()
    theme2 = theme1 + theme(axis_line_x=element_line())
    assert theme2 != theme1
    assert theme2.themeables != theme1.themeables
    assert theme2.rcParams == theme1.rcParams

    # specific difference
    for name in theme2.themeables:
        if name == "axis_line_x":
            assert theme2.themeables[name] != theme1.themeables[name]
        else:
            assert theme2.themeables[name] == theme1.themeables[name]


def test_add_partial_complete():
    theme1 = theme(axis_line_x=element_line())
    theme2 = theme_gray()
    theme3 = theme1 + theme2
    assert theme3 == theme2


def test_add_empty_theme_element():
    # An empty theme element does not alter the theme
    theme1 = theme_gray() + theme(axis_line_x=element_line(color="red"))
    theme2 = theme1 + theme(axis_line_x=element_line())
    assert theme1 == theme2


l1 = element_line(color="red", size=1, linewidth=1, linetype="solid")
l2 = element_line(color="blue", size=2, linewidth=2)
l3 = element_line(color="blue", size=2, linewidth=2, linetype="solid")
blank = element_blank()


def test_add_element_heirarchy():
    # parent themeable modifies child themeable
    theme1 = theme_gray() + theme(axis_line_x=l1)  # child
    theme2 = theme1 + theme(axis_line=l2)  # parent
    theme3 = theme1 + theme(axis_line_x=l3)  # child, for comparison
    assert theme2.themeables["axis_line_x"] == theme3.themeables["axis_line_x"]

    theme1 = theme_gray() + theme(axis_line_x=l1)  # child
    theme2 = theme1 + theme(line=l2)  # grand-parent
    theme3 = theme1 + theme(axis_line_x=l3)  # child, for comparison
    assert theme2.themeables["axis_line_x"] == theme3.themeables["axis_line_x"]

    # child themeable does not affect parent
    theme1 = theme_gray() + theme(axis_line=l1)  # parent
    theme2 = theme1 + theme(axis_line_x=l2)  # child
    theme3 = theme1 + theme(axis_line=l3)  # parent, for comparison
    assert theme3.themeables["axis_line"] != theme2.themeables["axis_line"]


def test_add_element_blank():
    # Adding onto a blanked themeable
    theme1 = theme_gray() + theme(axis_line_x=l1)  # not blank
    theme2 = theme1 + theme(axis_line_x=blank)  # blank
    theme3 = theme2 + theme(axis_line_x=l3)  # not blank
    theme4 = theme_gray() + theme(axis_line_x=l3)  # for comparison
    assert theme3 != theme1
    assert theme3 != theme2
    assert theme3 == theme4  # blanking cleans the slate

    # When a themeable is blanked, the apply_ax method
    # is replaced with the blank method.
    th2 = theme2.themeables["axis_line_x"]
    th3 = theme3.themeables["axis_line_x"]
    assert th2.apply_ax.__name__ == "blank_ax"
    assert th3.apply_ax.__name__ == "apply_ax"


def test_element_line_dashed_capstyle():
    p = ggplot(mtcars, aes(x="wt", y="mpg")) + theme(
        panel_grid=element_line(
            linetype="dashed",
            lineend="butt",
            size=1.0,
        )
    )
    # no exception
    p._build()


def test_element_line_solid_capstyle():
    p = ggplot(mtcars, aes(x="wt", y="mpg")) + theme(
        panel_grid=element_line(
            linetype="solid",
            lineend="butt",
            size=1.0,
        )
    )
    # no exception
    p._build()


class TestThemes:
    g = (
        ggplot(mtcars, aes(x="wt", y="mpg", color="factor(gear)"))
        + geom_point()
        + facet_grid("vs ~ am")
    )

    def test_theme_538(self):
        p = self.g + labs(title="Theme 538") + theme_538()

        assert p == "theme_538"

    def test_theme_bw(self):
        p = self.g + labs(title="Theme BW") + theme_bw()

        assert p == "theme_bw"

    def test_theme_classic(self):
        p = self.g + labs(title="Theme Classic") + theme_classic()

        assert p == "theme_classic"

    def test_theme_dark(self):
        p = self.g + labs(title="Theme Dark") + theme_dark()

        assert p == "theme_dark"

    def test_theme_gray(self):
        p = self.g + labs(title="Theme Gray") + theme_gray()

        assert p == "theme_gray"

    def test_theme_light(self):
        p = self.g + labs(title="Theme Light") + theme_light()

        assert p == "theme_light"

    def test_theme_linedraw(self):
        p = self.g + labs(title="Theme Linedraw") + theme_linedraw()

        assert p == "theme_linedraw"

    def test_theme_matplotlib(self):
        p = self.g + labs(title="Theme Matplotlib") + theme_matplotlib()

        assert p == "theme_matplotlib"

    def test_theme_minimal(self):
        p = self.g + labs(title="Theme Minimal") + theme_minimal()

        assert p == "theme_minimal"

    def test_theme_tufte(self):
        p = self.g + labs(title="Theme Tufte") + theme_tufte(ticks=False)

        assert p == "theme_tufte"

    @pytest.mark.xfail(reason="fails on github actions")
    def test_theme_seaborn(self):
        p = self.g + labs(title="Theme Seaborn") + theme_seaborn()

        assert p == "theme_seaborn"

    def test_theme_void(self):
        p = self.g + labs(title="Theme Void") + theme_void()

        assert p == "theme_void"

    def test_theme_xkcd(self):
        p = (
            self.g
            + labs(title="Theme Xkcd")
            + theme_xkcd()
            # High likely hood of Comic Sans being available
            + theme(text=element_text(family=["Comic Sans MS"]))
        )

        if os.environ.get("CI") or os.environ.get("TRAVIS"):
            # Github Actions and Travis do not have the fonts,
            # we still check to catch any other errors
            assert p != "theme_gray"
        else:
            assert p == "theme_xkcd"


class TestLayout:
    g = (
        ggplot(mtcars, aes(x="wt", y="mpg", color="factor(gear)"))
        + geom_point()
        + labs(  # New
            x="Weight",
            y="Miles Per Gallon",
            title="Relationship between Weight and Fuel Efficiency in Cars",
            subtitle="Should we be driving lighter cars?",
            caption=(
                "The plot shows a negative correlation between car weight\n"
                "and fuel efficiency, with lighter cars generally achieving\n"
                "higher miles per gallon"
            ),
        )
    )

    def test_default(self):
        assert self.g == "default"

    def test_legend_on_top(self):
        p = self.g + theme(legend_position="top")
        assert p == "legend_at_top"

    def test_legend_on_the_left(self):
        p = self.g + theme(legend_position="left")
        assert p == "legend_on_the_left"

    def test_legend_at_the_bottom(self):
        p = self.g + theme(legend_position="bottom")
        assert p == "legend_at_the_bottom"

    def test_facet_grid(self):
        p = self.g + facet_grid("am ~ gear")
        assert p == "facet_grid"

    def test_facet_wrap(self):
        p = self.g + facet_wrap("carb", nrow=2)
        assert p == "facet_wrap"

    def test_facet_wrap_scales_free(self):
        p = self.g + facet_wrap("carb", scales="free")
        assert p == "facet_wrap_scales_free"



from __future__ import annotations

import typing
from contextlib import suppress
from warnings import warn

import numpy as np

from ..doctools import document
from ..exceptions import PlotnineError, PlotnineWarning
from ..positions import position_nudge
from ..utils import order_as_data_mapping, to_rgba
from .geom import geom

if typing.TYPE_CHECKING:
    from typing import Any

    import pandas as pd

    from plotnine.iapi import panel_view
    from plotnine.typing import Aes, Axes, Coord, DataLike, DrawingArea, Layer


# Note: hjust & vjust are parameters instead of aesthetics
# due to a limitation imposed by MPL
# see: https://github.com/matplotlib/matplotlib/pull/1181
@document
class geom_text(geom):
    """
    Textual annotations

    {usage}

    Parameters
    ----------
    {common_parameters}
    parse : bool (default: False)
        If :py:`True`, the labels will be rendered with
        `latex <http://matplotlib.org/users/usetex.html>`_.
    family : str (default: None)
        Font family.
    fontweight : int or str (default: normal)
        Font weight.
    fontstyle : str (default: normal)
        Font style. One of *normal*, *italic* or *oblique*
    nudge_x : float (default: 0)
        Horizontal adjustment to apply to the text
    nudge_y : float (default: 0)
        Vertical adjustment to apply to the text
    adjust_text: dict (default: None)
        Parameters to :class:`adjustText.adjust_text` will repel
        overlapping texts. This parameter takes priority of over
        ``nudge_x`` and ``nudge_y``.

        ``adjust_text`` does not work well when it is used in the
        first layer of the plot, or if it is the only layer.
        For more see the documentation at
        https://github.com/Phlya/adjustText/wiki .
    format_string : str (default: None)
        If not :py:`None`, then the text is formatted with this
        string using :meth:`str.format` e.g::

            # 2.348 -> "2.35%"
            geom_text(format_string="{:.2f}%")

    path_effects : list (default: None)
        If not :py:`None`, then the text will use these effects.
        See `path_effects
        <https://matplotlib.org/tutorials/advanced/patheffects_guide.html>`_
        documentation for more details.

    See Also
    --------
    plotnine.geoms.geom_label
    matplotlib.text.Text
    matplotlib.patheffects

    """

    _aesthetics_doc = """
    {aesthetics_table}

    .. rubric:: Aesthetics Descriptions

    ha
        Horizontal alignment. One of *left*, *center* or *right.*

    va
        Vertical alignment. One of *top*, *center*, *bottom*, *baseline*.

    """
    DEFAULT_AES = {
        "alpha": 1,
        "angle": 0,
        "color": "black",
        "size": 11,
        "lineheight": 1.2,
        "ha": "center",
        "va": "center",
    }
    REQUIRED_AES = {"label", "x", "y"}
    DEFAULT_PARAMS = {
        "stat": "identity",
        "position": "identity",
        "na_rm": False,
        "parse": False,
        "family": None,
        "fontweight": "normal",
        "fontstyle": "normal",
        "nudge_x": 0,
        "nudge_y": 0,
        "adjust_text": None,
        "format_string": None,
        "path_effects": None,
    }

    def __init__(
        self,
        mapping: Aes | None = None,
        data: DataLike | None = None,
        **kwargs: Any,
    ):
        data, mapping = order_as_data_mapping(data, mapping)
        nudge_kwargs = {}
        adjust_text = kwargs.get("adjust_text", None)
        if adjust_text is None:
            with suppress(KeyError):
                nudge_kwargs["x"] = kwargs["nudge_x"]
            with suppress(KeyError):
                nudge_kwargs["y"] = kwargs["nudge_y"]
            if nudge_kwargs:
                kwargs["position"] = position_nudge(**nudge_kwargs)
        else:
            check_adjust_text()

        # Accomodate the old names
        if mapping and "hjust" in mapping:
            mapping["ha"] = mapping.pop("hjust")

        if mapping and "vjust" in mapping:
            mapping["va"] = mapping.pop("vjust")

        geom.__init__(self, mapping, data, **kwargs)

    def setup_data(self, data: pd.DataFrame) -> pd.DataFrame:
        parse = self.params["parse"]
        fmt = self.params["format_string"]

        def _format(series: pd.Series, tpl: str) -> list[str | None]:
            """
            Format items in series

            Missing values are preserved as None
            """
            if series.dtype == float:
                return [None if np.isnan(l) else tpl.format(l) for l in series]
            else:
                return [None if l is None else tpl.format(l) for l in series]

        # format
        if fmt:
            data["label"] = _format(data["label"], fmt)

        # Parse latex
        if parse:
            data["label"] = _format(data["label"], "${}$")

        return data

    def draw_panel(
        self,
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        super().draw_panel(data, panel_params, coord, ax, **params)

    @staticmethod
    def draw_group(
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        data = coord.transform(data, panel_params)

        # Bind color and alpha
        color = to_rgba(data["color"], data["alpha"])

        # Create a dataframe for the plotting data required
        # by ax.text
        df = data[["x", "y", "size"]].copy()
        df["s"] = data["label"]
        df["rotation"] = data["angle"]
        df["linespacing"] = data["lineheight"]
        df["color"] = color
        df["ha"] = data["ha"]
        df["va"] = data["va"]
        df["family"] = params["family"]
        df["fontweight"] = params["fontweight"]
        df["fontstyle"] = params["fontstyle"]
        df["zorder"] = params["zorder"]
        df["rasterized"] = params["raster"]
        df["clip_on"] = True

        # 'boxstyle' indicates geom_label so we need an MPL bbox
        draw_label = "boxstyle" in params
        if draw_label:
            fill = to_rgba(data.pop("fill"), data["alpha"])
            if isinstance(fill, tuple):
                fill = [list(fill)] * len(data["x"])
            df["facecolor"] = fill

            if params["boxstyle"] in ("round", "round4"):
                boxstyle = "{},pad={},rounding_size={}".format(
                    params["boxstyle"],
                    params["label_padding"],
                    params["label_r"],
                )
            elif params["boxstyle"] in ("roundtooth", "sawtooth"):
                boxstyle = "{},pad={},tooth_size={}".format(
                    params["boxstyle"],
                    params["label_padding"],
                    params["tooth_size"],
                )
            else:
                boxstyle = "{},pad={}".format(
                    params["boxstyle"], params["label_padding"]
                )
            bbox = {"linewidth": params["label_size"], "boxstyle": boxstyle}
        else:
            bbox = {}

        texts = []

        # For labels add a bbox
        for i in range(len(data)):
            kw: dict["str", Any] = df.iloc[i].to_dict()
            if draw_label:
                kw["bbox"] = bbox
                kw["bbox"]["edgecolor"] = params["boxcolor"] or kw["color"]
                kw["bbox"]["facecolor"] = kw.pop("facecolor")
            text_elem = ax.text(**kw)
            texts.append(text_elem)
            if params["path_effects"]:
                text_elem.set_path_effects(params["path_effects"])

        _adjust = params["adjust_text"]
        if _adjust:
            from adjustText import adjust_text

            if params["zorder"] == 1:
                warn(
                    "For better results with adjust_text, it should "
                    "not be the first layer or the only layer.",
                    PlotnineWarning,
                )

            arrowprops = _adjust.pop("arrowprops", {})
            if "color" not in arrowprops:
                arrowprops["color"] = color[0]

            adjust_text(texts, ax=ax, arrowprops=arrowprops, **_adjust)

    @staticmethod
    def draw_legend(
        data: pd.Series[Any], da: DrawingArea, lyr: Layer
    ) -> DrawingArea:
        """
        Draw letter 'a' in the box

        Parameters
        ----------
        data : Series
            Data Row
        da : DrawingArea
            Canvas
        lyr : layer
            Layer

        Returns
        -------
        out : DrawingArea
        """
        from matplotlib.text import Text

        color = to_rgba(data["color"], data["alpha"])

        key = Text(
            x=0.5 * da.width,  # pyright: ignore[reportGeneralTypeIssues]
            y=0.5 * da.height,  # pyright: ignore[reportGeneralTypeIssues]
            text="a",
            size=data["size"],
            family=lyr.geom.params["family"],
            color=color,
            rotation=data["angle"],
            horizontalalignment="center",
            verticalalignment="center",
        )
        da.add_artist(key)
        return da


def check_adjust_text():
    try:
        pass
    except ImportError as err:
        raise PlotnineError(
            "To use adjust_text you must install the adjustText package."
        ) from err



from ..doctools import document
from ..mapping.aes import ALL_AESTHETICS
from ..mapping.evaluation import after_stat
from ..utils import groupby_apply
from .stat import stat


@document
class stat_sum(stat):
    """
    Sum unique values

    Useful for overplotting on scatterplots.

    {usage}

    Parameters
    ----------
    {common_parameters}
    """

    _aesthetics_doc = """
    {aesthetics_table}

    .. rubric:: Options for computed aesthetics

    ::

        'n'     # Number of observations at a position
        'prop'  # Ratio of points in that panel at a position

    """

    REQUIRED_AES = {"x", "y"}
    DEFAULT_PARAMS = {"geom": "point", "position": "identity", "na_rm": False}
    DEFAULT_AES = {"size": after_stat("n"), "weight": 1}
    CREATES = {"n", "prop"}

    @classmethod
    def compute_panel(cls, data, scales, **params):
        if "weight" not in data:
            data["weight"] = 1

        def count(df):
            """
            Do a weighted count
            """
            df["n"] = df["weight"].sum()
            return df.iloc[0:1]

        def ave(df):
            """
            Calculate proportion values
            """
            df["prop"] = df["n"] / df["n"].sum()
            return df

        # group by all present aesthetics other than the weight,
        # then sum them (i.e no. of uniques) to get the raw count
        # 'n', and the proportions 'prop' per group
        s: set[str] = set(data.columns) & ALL_AESTHETICS
        by = list(s.difference(["weight"]))
        counts = groupby_apply(data, by, count)
        counts = groupby_apply(counts, "group", ave)
        return counts



from __future__ import annotations

import typing

import numpy as np

from ..doctools import document
from .geom_segment import geom_segment

if typing.TYPE_CHECKING:
    import pandas as pd


@document
class geom_spoke(geom_segment):
    """
    Line segment parameterised by location, direction and distance

    {usage}

    Parameters
    ----------
    {common_parameters}

    See Also
    --------
    plotnine.geoms.geom_segment : For documentation of extra
        parameters.
    """

    REQUIRED_AES = {"x", "y", "angle", "radius"}

    def setup_data(self, data: pd.DataFrame) -> pd.DataFrame:
        try:
            radius = data["radius"]
        except KeyError:
            radius = self.aes_params["radius"]
        try:
            angle = data["angle"]
        except KeyError:
            angle = self.aes_params["angle"]

        data["xend"] = data["x"] + np.cos(angle) * radius
        data["yend"] = data["y"] + np.sin(angle) * radius
        return data



from __future__ import annotations

import typing
from contextlib import suppress

import numpy as np
import pandas as pd

from ..exceptions import PlotnineError
from ..iapi import labels_view, layout_details, pos_scales
from ..utils import match

if typing.TYPE_CHECKING:
    from plotnine.iapi import panel_view
    from plotnine.typing import (
        Axes,
        Coord,
        Facet,
        Ggplot,
        Layers,
        Scales,
    )


class Layout:
    """
    Layout of entire plot
    """

    #: facet
    facet: Facet

    #: coordinate system
    coord: Coord

    #: A dataframe with the layout information of the plot
    layout: pd.DataFrame

    #: List of x scales
    panel_scales_x: Scales

    #: List of y scales
    panel_scales_y: Scales

    #: Range & breaks information for each panel
    panel_params: list[panel_view]

    axs: list[Axes]  # MPL axes

    def setup(self, layers: Layers, plot: Ggplot):
        """
        Create a layout for the panels

        The layout is a dataframe that stores all the
        structual information about the panels that will
        make up the plot. The actual layout depends on
        the type of facet.

        This method ensures that each layer has a copy of the
        data it needs in `layer.data`. That data is also has
        column `PANEL` that indicates the panel onto which each
        data row/item will be plotted.
        """
        data = [l.data for l in layers]

        # setup facets
        self.facet = plot.facet
        self.facet.setup_params(data)
        data = self.facet.setup_data(data)

        # setup coords
        self.coord = plot.coordinates
        self.coord.setup_params(data)
        data = self.coord.setup_data(data)

        # Generate panel layout
        data = self.facet.setup_data(data)
        self.layout = self.facet.compute_layout(data)
        self.layout = self.coord.setup_layout(self.layout)
        self.check_layout()

        # Map the data to the panels
        for layer, ldata in zip(layers, data):
            layer.data = self.facet.map(ldata, self.layout)

    def train_position(self, layers: Layers, scales: Scales):
        """
        Create all the required x & y panel_scales

        And set the ranges for each scale according to the data

        Notes
        -----
        The number of x or y scales depends on the facetting,
        particularly the scales parameter. e.g if `scales='free'`
        then each panel will have separate x and y scales, and
        if `scales='fixed'` then all panels will share an x
        scale and a y scale.
        """
        layout = self.layout
        if not hasattr(self, "panel_scales_x") and scales.x:
            result = self.facet.init_scales(layout, scales.x, None)
            self.panel_scales_x = result.x

        if not hasattr(self, "panel_scales_y") and scales.y:
            result = self.facet.init_scales(layout, None, scales.y)
            self.panel_scales_y = result.y

        self.facet.train_position_scales(self, layers)

    def map_position(self, layers: Layers):
        """
        Map x & y (position) aesthetics onto the scales.

        e.g If the x scale is scale_x_log10, after this
        function all x, xmax, xmin, ... columns in data
        will be mapped onto log10 scale (log10 transformed).
        The real mapping is handled by the scale.map
        """
        _layout = self.layout

        for layer in layers:
            data = layer.data
            match_id = match(data["PANEL"], _layout["PANEL"])
            if self.panel_scales_x:
                x_vars = list(
                    set(self.panel_scales_x[0].aesthetics) & set(data.columns)
                )
                SCALE_X = _layout["SCALE_X"].iloc[match_id].tolist()
                self.panel_scales_x.map(data, x_vars, SCALE_X)

            if self.panel_scales_y:
                y_vars = list(
                    set(self.panel_scales_y[0].aesthetics) & set(data.columns)
                )
                SCALE_Y = _layout["SCALE_Y"].iloc[match_id].tolist()
                self.panel_scales_y.map(data, y_vars, SCALE_Y)

    def get_scales(self, i: int) -> pos_scales:
        """
        Return x & y scales for panel i

        Parameters
        ----------
        i : int
          Panel id

        Returns
        -------
        scales : types.SimpleNamespace
          Class attributes *x* for the x scale and *y*
          for the y scale of the panel

        """
        # wrapping with np.asarray prevents an exception
        # on some datasets
        bool_idx = np.asarray(self.layout["PANEL"]) == i

        idx = self.layout["SCALE_X"].loc[bool_idx].iloc[0]
        xsc = self.panel_scales_x[idx - 1]

        idx = self.layout["SCALE_Y"].loc[bool_idx].iloc[0]
        ysc = self.panel_scales_y[idx - 1]

        return pos_scales(x=xsc, y=ysc)

    def reset_position_scales(self):
        """
        Reset x and y scales
        """
        if not self.facet.shrink:
            return

        with suppress(AttributeError):
            self.panel_scales_x.reset()

        with suppress(AttributeError):
            self.panel_scales_y.reset()

    def setup_panel_params(self, coord: Coord):
        """
        Calculate the x & y range & breaks information for each panel

        Parameters
        ----------
        coord : coord
            Coordinate
        """
        if not self.panel_scales_x:
            raise PlotnineError("Missing an x scale")

        if not self.panel_scales_y:
            raise PlotnineError("Missing a y scale")

        self.panel_params = []
        cols = ["SCALE_X", "SCALE_Y"]
        for i, j in self.layout[cols].itertuples(index=False):
            i, j = i - 1, j - 1
            params = coord.setup_panel_params(
                self.panel_scales_x[i], self.panel_scales_y[j]
            )
            self.panel_params.append(params)

    def finish_data(self, layers: Layers):
        """
        Modify data before it is drawn out by the geom

        Parameters
        ----------
        layers : list
            List of layers
        """
        for layer in layers:
            layer.data = self.facet.finish_data(layer.data, self)

    def check_layout(self):
        required = {"PANEL", "SCALE_X", "SCALE_Y"}
        common = self.layout.columns.intersection(list(required))
        if len(required) != len(common):
            raise PlotnineError(
                "Facet layout has bad format. It must contain "
                f"the columns '{required}'"
            )

    def xlabel(self, labels: labels_view) -> str:
        """
        Determine x-axis label

        Parameters
        ----------
        labels : labels_view
            Labels as specified by the user through the ``labs`` or
            ``xlab`` calls.

        Returns
        -------
        out : str
            x-axis label
        """
        if self.panel_scales_x[0].name is not None:
            return self.panel_scales_x[0].name
        elif labels.x is not None:
            return labels.x
        return ""

    def ylabel(self, labels: labels_view) -> str:
        """
        Determine y-axis label

        Parameters
        ----------
        labels : labels_view
            Labels as specified by the user through the ``labs`` or
            ``ylab`` calls.

        Returns
        -------
        out : str
            y-axis label
        """
        if self.panel_scales_y[0].name is not None:
            return self.panel_scales_y[0].name
        elif labels.y is not None:
            return labels.y
        return ""

    def set_xy_labels(self, labels: labels_view) -> labels_view:
        """
        Determine x & y axis labels

        Parameters
        ----------
        labels : labels_view
            Labels as specified by the user through the ``labs`` or
            ``ylab`` calls.

        Returns
        -------
        out : labels_view
            Modified labels
        """
        labels.x = self.xlabel(labels)
        labels.y = self.ylabel(labels)
        return labels

    def get_details(self) -> list[layout_details]:
        columns = [
            "PANEL",
            "ROW",
            "COL",
            "SCALE_X",
            "SCALE_Y",
            "AXIS_X",
            "AXIS_Y",
        ]
        vcols = self.layout.columns.difference(columns)
        lst = []
        for pidx, row in self.layout.iterrows():
            ld = layout_details(
                panel_index=pidx,  # type: ignore
                variables={name: row[name] for name in vcols},
                **{str.lower(k): row[k] for k in columns},
            )
            lst.append(ld)
        return lst



from __future__ import annotations

import typing

from .exceptions import PlotnineError
from .iapi import labels_view
from .mapping.aes import SCALED_AESTHETICS, rename_aesthetics

if typing.TYPE_CHECKING:
    import plotnine as p9

__all__ = ["xlab", "ylab", "labs", "ggtitle"]
VALID_LABELS = SCALED_AESTHETICS | {"caption", "title", "subtitle"}


class labs:
    """
    Add labels for aesthetics and/or title

    Parameters
    ----------
    kwargs : dict
        Aesthetics (with scales) to be renamed. You can also
        set the ``title`` and ``caption``.
    """

    labels: labels_view

    def __init__(self, **kwargs: str):
        unknown = kwargs.keys() - VALID_LABELS
        if unknown:
            raise PlotnineError(f"Cannot deal with these labels: {unknown}")
        self.labels = labels_view(**rename_aesthetics(kwargs))

    def __radd__(self, gg: p9.ggplot) -> p9.ggplot:
        """
        Add labels to ggplot object
        """
        gg.labels.update(self.labels)
        return gg


class xlab(labs):
    """
    Create x-axis label

    Parameters
    ----------
    xlab : str
        x-axis label
    """

    def __init__(self, xlab: str):
        if xlab is None:
            raise PlotnineError("Arguments to xlab cannot be None")
        self.labels = labels_view(x=xlab)


class ylab(labs):
    """
    Create y-axis label

    Parameters
    ----------
    ylab : str
        y-axis label
    """

    def __init__(self, ylab: str):
        if ylab is None:
            raise PlotnineError("Arguments to ylab cannot be None")
        self.labels = labels_view(y=ylab)


class ggtitle(labs):
    """
    Create plot title

    Parameters
    ----------
    title : str
        Plot title
    """

    def __init__(self, title: str):
        if title is None:
            raise PlotnineError("Arguments to ggtitle cannot be None")
        self.labels = labels_view(title=title)



import pandas as pd

from plotnine import aes, geom_dotplot, ggplot

df = pd.DataFrame({"x": [1, 2, 2, 3, 3, 3, 4, 4, 4, 4]})


def test_dotdensity():
    p = ggplot(df, aes("x")) + geom_dotplot(bins=15)

    assert p == "dotdensity"


def test_histodot():
    p = ggplot(df, aes("x")) + geom_dotplot(bins=15, method="histodot")

    assert p == "histodot"


def test_stackratio():
    p = ggplot(df, aes("x")) + geom_dotplot(bins=15, stackratio=0.75)

    assert p == "stackratio"


def test_binaxis_y():
    p = ggplot(df, aes("x", "x")) + geom_dotplot(bins=15, binaxis="y")

    assert p == "binaxis_y"


def test_stackdir_down():
    p = ggplot(df, aes("x")) + geom_dotplot(bins=15, stackdir="down")

    assert p == "stackdir_down"


def test_stackdir_center():
    p = ggplot(df, aes("x")) + geom_dotplot(bins=15, stackdir="center")

    assert p == "stackdir_center"


def test_stackdir_centerwhole():
    p = ggplot(df, aes("x")) + geom_dotplot(bins=15, stackdir="centerwhole")

    assert p == "stackdir_centerwhole"


class TestGrouping:
    df = pd.DataFrame(
        {
            "x": [1, 2, 2, 3, 3, 3, 4, 4, 4, 4],
            "g": [1, 2, 22, 3, 33, 333, 4, 44, 444, 4444],
        }
    )

    p = ggplot(df, aes("x", "x", fill="factor(g)", color="factor(g)"))

    def test_group_basic(self):
        p = self.p + geom_dotplot(bins=15)

        assert p == "group_basic"

    def test_group_binpositions_all(self):
        p = self.p + geom_dotplot(bins=15, binpositions="all")

        assert p == "group_basic"  # Yes same as above

    def test_group_stackgroups(self):
        p = self.p + geom_dotplot(
            bins=15, binpositions="all", stackgroups=True
        )

        assert p == "group_stackgroups"

    def test_group_stackgroups_binaxis_y(self):
        p = self.p + geom_dotplot(
            bins=15, binpositions="all", stackgroups=True, binaxis="y"
        )

        assert p == "group_stackgroups_binaxis_y"



import numpy as np
import pandas as pd

from plotnine import aes, geom_qq, geom_qq_line, ggplot

random_state = np.random.RandomState(1234567890)
df_normal = pd.DataFrame({"x": random_state.normal(size=100)})


def test_normal():
    p = ggplot(df_normal, aes(sample="x")) + geom_qq()
    # Roughly a straight line of points through the origin
    assert p == "normal"


def test_normal_with_line():
    p = ggplot(df_normal, aes(sample="x")) + geom_qq() + geom_qq_line()
    # Roughly a straight line of points through the origin
    assert p == "normal_with_line"



import numpy as np
import pandas as pd
import pytest
import statsmodels.api as sm

from plotnine import (
    aes,
    coord_trans,
    geom_point,
    geom_smooth,
    ggplot,
    stat_smooth,
)
from plotnine.exceptions import PlotnineWarning

random_state = np.random.RandomState(1234567890)
n = 100

# linear relationship
x = np.linspace(0, 1, n)
y = 4 * x + 5
y_noisy = y + 0.1 * random_state.randn(n)
df_linear = pd.DataFrame({"x": x, "y": y, "y_noisy": y_noisy})


# non-linear relationship
x = np.linspace(-2 * np.pi, 2 * np.pi, n)
y = np.sin(x)
y_noisy = y + 0.1 * random_state.randn(n)
df_non_linear = pd.DataFrame({"x": x, "y": y, "y_noisy": y_noisy})

# discrete_x
df_discrete_x = pd.DataFrame(
    {"x": range(10), "y": [1, 2, 3, 4, 4, 5, 6, 7, 8, 9]}
)

# continuous_x
df_continuous_x = pd.DataFrame(
    {"x": np.arange(1, 21) + 0.2, "y": range(1, 21)}
)

# linear relationship, values greater than zero
n = 10
x = np.arange(1, 1 + n)
y = x + 11
y_noisy = y + random_state.rand(n)
df_linear_gtz = pd.DataFrame({"x": x, "y": y, "y_noisy": y_noisy})


def test_linear_smooth():
    p = (
        ggplot(df_linear, aes("x"))
        + geom_point(aes(y="y_noisy"))
        + geom_smooth(aes(y="y_noisy"), method="lm", span=0.3, color="blue")
    )

    assert p == "linear_smooth"


def test_linear_smooth_no_ci():
    p = (
        ggplot(df_linear, aes("x"))
        + geom_point(aes(y="y_noisy"))
        + geom_smooth(
            aes(y="y_noisy"), method="lm", span=0.3, color="blue", se=False
        )
    )

    assert p == "linear_smooth_no_ci"


def test_non_linear_smooth():
    p = (
        ggplot(df_linear, aes("x"))
        + geom_point(aes(y="y_noisy"))
        + geom_smooth(aes(y="y_noisy"), method="loess", span=0.3, color="blue")
    )

    assert p == "non_linear_smooth"


def test_non_linear_smooth_no_ci():
    p = (
        ggplot(df_linear, aes("x"))
        + geom_point(aes(y="y_noisy"))
        + geom_smooth(
            aes(y="y_noisy"), method="loess", span=0.3, color="blue", se=False
        )
    )

    assert p == "non_linear_smooth_no_ci"


def test_discrete_x():
    p = (
        ggplot(df_discrete_x, aes("x", "y"))
        + geom_point()
        + geom_smooth(color="blue")
    )

    assert p == "discrete_x"


def test_discrete_x_fullrange():
    p = (
        ggplot(df_discrete_x, aes("x", "y"))
        + geom_point()
        + geom_smooth(color="blue", fullrange=True)
    )

    assert p == "discrete_x_fullrange"


def test_continuous_x():
    n = len(df_continuous_x)
    p = (
        ggplot(df_continuous_x, aes("x", "y"))
        + geom_point()
        + geom_smooth(
            df_continuous_x[3 : n - 3],
            method="loess",
            color="blue",
            fullrange=False,
        )
    )
    assert p == "continuous_x"


def test_continuous_x_fullrange():
    n = len(df_continuous_x)
    p = (
        ggplot(df_continuous_x, aes("x", "y"))
        + geom_point()
        + geom_smooth(
            df_continuous_x[3 : n - 3],
            method="loess",
            color="blue",
            fullrange=True,
            method_args={"surface": "direct"},
        )
    )

    assert p == "continuous_x_fullrange"


def test_coord_trans_se_false():
    # scatter plot with LM fit using log-log coordinates
    p = (
        ggplot(df_linear_gtz, aes(x="x", y="y_noisy"))
        + geom_point()
        + coord_trans(x="log10", y="log10")
        + geom_smooth(method="lm", se=False)
    )
    assert p == "coord_trans_se_false"


class TestOther:
    p = ggplot(df_linear, aes("x")) + geom_point(aes(y="y_noisy"))

    def test_wls(self):
        p = self.p + geom_smooth(aes(y="y_noisy"), method="wls")
        p.draw_test()

    def test_rlm(self):
        p = self.p + geom_smooth(aes(y="y_noisy"), method="rlm")
        with pytest.warns(PlotnineWarning):
            p.draw_test()

    def test_glm(self):
        p = self.p + geom_smooth(aes(y="y_noisy"), method="glm")
        p.draw_test()

    def test_gls(self):
        p = self.p + geom_smooth(aes(y="y_noisy"), method="gls")
        p.draw_test()

    def test_lowess(self):
        p = self.p + geom_smooth(aes(y="y_noisy"), method="lowess")
        with pytest.warns(PlotnineWarning):
            p.draw_test()

    def test_mavg(self):
        p = self.p + geom_smooth(
            aes(y="y_noisy"), method="mavg", method_args={"window": 10}
        )
        p.draw_test()

    def test_gpr(self):
        try:
            from sklearn import gaussian_process  # noqa:401
        except ImportError:
            return

        p = self.p + geom_smooth(aes(y="y_noisy"), method="gpr")
        with pytest.warns(UserWarning):
            p.draw_test()


def test_sorts_by_x():
    df = pd.DataFrame({"x": [5, 0, 1, 2, 3, 4], "y": range(6)})
    p = ggplot(df, aes("x", "y")) + geom_smooth(stat="identity")

    assert p == "sorts_by_x"


def test_legend_fill_ratio():
    p = (
        ggplot(df_linear, aes("x", color="x<0.5"))
        + geom_point(aes(y="y_noisy"))
        + geom_smooth(aes(y="y_noisy"), method="lm", size=0.5, span=0.3)
    )

    assert p == "legend_fill_ratio"


def test_init_and_fit_kwargs():
    df = pd.DataFrame(
        {
            "x": np.arange(11),
            "y": [0, 0, 0, 0.05, 0.25, 0.5, 0.75, 0.95, 1, 1, 1],
        }
    )

    p = (
        ggplot(df, aes("x", "y"))
        + geom_point()
        + geom_smooth(
            method="glm",
            method_args={
                "family": sm.families.Binomial(),  # init parameter
                "method": "minimize",  # fit parameter
            },
            se=False,
        )
    )

    assert p == "init_and_fit_kwargs"


n = 100
random_state = np.random.RandomState(123)
mu = 0
sigma = 0.065
noise = random_state.randn(n) * sigma + mu
x = np.linspace(-2 * np.pi, 2 * np.pi, n)
df = pd.DataFrame(
    {
        "x": x,
        "y": np.sin(x) + noise,
    }
)


class TestFormula:
    p = ggplot(df, aes("x", "y")) + geom_point()

    def test_lm(self):
        p = self.p + stat_smooth(
            method="lm", formula="y ~ np.sin(x)", fill="red", se=True
        )
        assert p == "lm_formula"

    def test_lm_weights(self):
        p = (
            self.p
            + aes(weight="x.abs()")
            + stat_smooth(
                method="lm", formula="y ~ np.sin(x)", fill="red", se=True
            )
        )
        assert p == "lm_formula_weights"

    def test_glm(self):
        p = self.p + stat_smooth(
            method="glm", formula="y ~ np.sin(x)", fill="red", se=True
        )
        assert p == "glm_formula"

    def test_rlm(self):
        p = self.p + stat_smooth(
            method="rlm", formula="y ~ np.sin(x)", fill="red", se=False
        )

        assert p == "rlm_formula"

    def test_gls(self):
        p = self.p + stat_smooth(
            method="gls", formula="y ~ np.sin(x)", fill="red", se=True
        )
        assert p == "gls_formula"



from .elements import element_blank, element_line, element_rect, element_text
from .theme import theme
from .theme_gray import theme_gray


class theme_light(theme_gray):
    """
    A theme similar to :class:`theme_linedraw`

    Has light grey lines lines and axes to direct more attention
    towards the data.

    Parameters
    ----------
    base_size : int, optional
        Base font size. All text sizes are a scaled versions of
        the base font size. Default is 11.
    base_family : str, optional
        Base font family.
    """

    def __init__(self, base_size=11, base_family=None):
        super().__init__(base_size, base_family)
        self += theme(
            axis_ticks=element_line(color="#B3B3B3", size=0.5),
            axis_ticks_minor=element_blank(),
            legend_key=element_rect(fill="white", color="#7F7F7F", size=0.5),
            panel_background=element_rect(fill="white"),
            panel_border=element_rect(fill="None", color="#B3B3B3", size=1),
            panel_grid_major=element_line(color="#D9D9D9", size=0.5),
            panel_grid_minor=element_line(color="#EDEDED", size=0.25),
            strip_background=element_rect(
                fill="#B3B3B3", color="#B3B3B3", size=1
            ),
            strip_text_x=element_text(color="white"),
            strip_text_y=element_text(color="white", angle=-90),
        )



from plotnine import position_stack
from plotnine.doctools import document
from plotnine.geoms.geom import geom
from plotnine.scales.scale import scale
from plotnine.stats.stat import stat


@document
class scale_expand(scale):
    """
    Expand

    Parameters
    ----------
    base_param_1 : int or float
        Base Param 1 Description
    base_param_2 : str
        Base Param 2 Description
    base_param_3 : dict
        Base Param 3 Description
    specific_parameter : str, optional
        Base Specific Parameter Description
    """


class mixin:
    """
    Expand

    Parameters
    ----------
    mixin_param_1 : int or float
        Mixin Param 1 Description
    mixin_param_2 : str
        Mixin Param 2 Description
    """


@document
class scale_expand_earth(mixin, scale_expand):
    """
    Expand Earth

    Parameters
    ----------
    {superclass_parameters}
    derived_param_1 : int or float
        Derived Param 1 Description
    derived_param_2: float
        Derived Param 2 Description
    specific_parameter : str
        Derived Specific Parameter Description
    """


def test_document_scale():
    doc = scale_expand_earth.__doc__
    assert doc.count("base_param_1") == 1
    assert doc.count("base_param_2") == 1
    assert doc.count("base_param_3") == 1
    assert doc.count("derived_param_1") == 1
    assert doc.count("derived_param_2") == 1
    assert doc.count("mixin_param_1") == 1
    assert doc.count("mixin_param_2") == 1

    # overridden parameter
    assert "specific_parameter : str, optional" not in doc
    assert "Base Specific Parameter Description" not in doc
    assert doc.count("specific_parameter : str") == 1
    assert doc.count("Derived Specific Parameter Description") == 1


@document
class geom_abc(geom):
    """
    Geom ABC

    {usage}

    Parameters
    ----------
    {common_parameters}
    """

    DEFAULT_AES = {"color": None}
    DEFAULT_PARAMS = {
        "stat": "bin",
        "position": position_stack,
        "na_rm": False,
    }


def test_document_geom():
    doc = geom_abc.__doc__
    assert "~plotnine.stats.stat_bin" in doc
    assert "position_stack" in doc


@document
class stat_abc(stat):
    """
    Stat ABC

    {usage}

    Parameters
    ----------
    {common_parameters}
    """

    DEFAULT_AES = {"weight": None}
    DEFAULT_PARAMS = {"geom": geom_abc, "position": "stack", "na_rm": False}


def test_document_stat():
    doc = stat_abc.__doc__
    assert "geom_abc" in doc
    assert "~plotnine.positions.position_stack" in doc



from __future__ import annotations

import typing
from collections import Counter
from contextlib import suppress
from warnings import warn

import numpy as np

from ..doctools import document
from ..exceptions import PlotnineWarning
from ..utils import SIZE_FACTOR, make_line_segments, match, to_rgba
from .geom import geom

if typing.TYPE_CHECKING:
    from typing import Any, Literal, Sequence

    import numpy.typing as npt
    import pandas as pd
    from matplotlib.path import Path

    from plotnine.iapi import panel_view
    from plotnine.typing import Axes, Coord, DrawingArea, Layer, TupleFloat2


@document
class geom_path(geom):
    """
    Connected points

    {usage}

    Parameters
    ----------
    {common_parameters}
    lineend : str (default: butt)
        Line end style, of of *butt*, *round* or *projecting.*
        This option is applied for solid linetypes.
    linejoin : str (default: round)
        Line join style, one of *round*, *miter* or *bevel*.
        This option is applied for solid linetypes.
    arrow : plotnine.geoms.geom_path.arrow (default: None)
        Arrow specification. Default is no arrow.

    See Also
    --------
    plotnine.geoms.arrow : for adding arrowhead(s) to paths.
    """

    DEFAULT_AES = {
        "alpha": 1,
        "color": "black",
        "linetype": "solid",
        "size": 0.5,
    }

    REQUIRED_AES = {"x", "y"}
    DEFAULT_PARAMS = {
        "stat": "identity",
        "position": "identity",
        "na_rm": False,
        "lineend": "butt",
        "linejoin": "round",
        "arrow": None,
    }

    def handle_na(self, data: pd.DataFrame) -> pd.DataFrame:
        def keep(x: Sequence[float]) -> npt.NDArray[np.bool_]:
            # first non-missing to last non-missing
            first = match([False], x, nomatch=1, start=0)[0]
            last = len(x) - match([False], x[::-1], nomatch=1, start=0)[0]
            bool_idx = np.hstack(
                [
                    np.repeat(False, first),
                    np.repeat(True, last - first),
                    np.repeat(False, len(x) - last),
                ]
            )
            return bool_idx

        # Get indices where any row for the select aesthetics has
        # NaNs at the beginning or the end. Those we drop
        bool_idx = (
            data[["x", "y", "size", "color", "linetype"]]
            .isnull()  # Missing
            .apply(keep, axis=0)
        )  # Beginning or the End
        bool_idx = np.all(bool_idx, axis=1)  # Across the aesthetics

        # return data
        n1 = len(data)
        data = data[bool_idx]
        data.reset_index(drop=True, inplace=True)
        n2 = len(data)

        if n2 != n1 and not self.params["na_rm"]:
            msg = "geom_path: Removed {} rows containing missing values."
            warn(msg.format(n1 - n2), PlotnineWarning)

        return data

    def draw_panel(
        self,
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        if not any(data["group"].duplicated()):
            warn(
                "geom_path: Each group consist of only one "
                "observation. Do you need to adjust the "
                "group aesthetic?",
                PlotnineWarning,
            )

        # drop lines with less than two points
        c = Counter(data["group"])
        counts = np.array([c[v] for v in data["group"]])
        data = data[counts >= 2]

        if len(data) < 2:
            return

        # dataframe mergesort is stable, we rely on that here
        data = data.sort_values("group", kind="mergesort")
        data.reset_index(drop=True, inplace=True)

        # When the parameters of the path are not constant
        # with in the group, then the lines that make the paths
        # can be drawn as separate segments
        cols = {"color", "size", "linetype", "alpha", "group"}
        cols = cols & set(data.columns)
        df = data.drop_duplicates(cols)
        constant = len(df) == data["group"].nunique()
        params["constant"] = constant

        if not constant:
            self.draw_group(data, panel_params, coord, ax, **params)
        else:
            for _, gdata in data.groupby("group"):
                gdata.reset_index(inplace=True, drop=True)
                self.draw_group(gdata, panel_params, coord, ax, **params)

    @staticmethod
    def draw_group(
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        data = coord.transform(data, panel_params, munch=True)
        data["size"] *= SIZE_FACTOR
        constant = params.pop("constant", data["group"].nunique() == 1)

        if not constant:
            _draw_segments(data, ax, **params)
        else:
            _draw_lines(data, ax, **params)

        if "arrow" in params and params["arrow"]:
            params["arrow"].draw(
                data, panel_params, coord, ax, constant=constant, **params
            )

    @staticmethod
    def draw_legend(
        data: pd.Series[Any], da: DrawingArea, lyr: Layer
    ) -> DrawingArea:
        """
        Draw a horizontal line in the box

        Parameters
        ----------
        data : Series
            Data Row
        da : DrawingArea
            Canvas
        lyr : layer
            Layer

        Returns
        -------
        out : DrawingArea
        """
        from matplotlib.lines import Line2D

        data["size"] *= SIZE_FACTOR
        x = [0, da.width]
        y = [0.5 * da.height] * 2
        color = to_rgba(data["color"], data["alpha"])

        key = Line2D(
            x,
            y,
            linestyle=data["linetype"],
            linewidth=data["size"],
            color=color,
            solid_capstyle="butt",
            antialiased=False,
        )
        da.add_artist(key)
        return da


class arrow:
    """
    Define arrow (actually an arrowhead)

    This is used to define arrow heads for
    :class:`.geom_path`.

    Parameters
    ----------
    angle : int | float
        angle in degrees between the tail a
        single edge.
    length : int | float
        of the edge in "inches"
    ends : str in ``['last', 'first', 'both']``
        At which end of the line to draw the
        arrowhead
    type : str in ``['open', 'closed']``
        When it is closed, it is also filled
    """

    def __init__(
        self,
        angle: float = 30,
        length: float = 0.2,
        ends: Literal["first", "last", "both"] = "last",
        type: Literal["open", "closed"] = "open",
    ):
        self.angle = angle
        self.length = length
        self.ends = ends
        self.type = type

    def draw(
        self,
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        constant: bool = True,
        **params: Any,
    ):
        """
        Draw arrows at the end(s) of the lines

        Parameters
        ----------
        data : dataframe
            Data to be plotted by this geom. This is the
            dataframe created in the plot_build pipeline.
        panel_params : panel_view
            The scale information as may be required by the
            axes. At this point, that information is about
            ranges, ticks and labels. Attributes are of interest
            to the geom are::

                'panel_params.x.range'  # tuple
                'panel_params.y.range'  # tuple

        coord : coord
            Coordinate (e.g. coord_cartesian) system of the
            geom.
        ax : axes
            Axes on which to plot.
        constant: bool
            If the path attributes vary along the way. If false,
            the arrows are per segment of the path
        params : dict
            Combined parameters for the geom and stat. Also
            includes the 'zorder'.
        """
        first = self.ends in ("first", "both")
        last = self.ends in ("last", "both")

        data = data.sort_values("group", kind="mergesort")
        data["color"] = to_rgba(data["color"], data["alpha"])

        if self.type == "open":
            data["facecolor"] = "none"
        else:
            data["facecolor"] = data["color"]

        if not constant:
            from matplotlib.collections import PathCollection

            # Get segments/points (x1, y1) -> (x2, y2)
            # for which to calculate the arrow heads
            idx1: list[int] = []
            idx2: list[int] = []
            for _, df in data.groupby("group"):
                idx1.extend(df.index[:-1].to_list())
                idx2.extend(df.index[1:].to_list())

            d = {
                "zorder": params["zorder"],
                "rasterized": params["raster"],
                "edgecolor": data.loc[idx1, "color"],
                "facecolor": data.loc[idx1, "facecolor"],
                "linewidth": data.loc[idx1, "size"],
                "linestyle": data.loc[idx1, "linetype"],
            }

            x1 = data.loc[idx1, "x"].to_numpy()
            y1 = data.loc[idx1, "y"].to_numpy()
            x2 = data.loc[idx2, "x"].to_numpy()
            y2 = data.loc[idx2, "y"].to_numpy()

            if first:
                paths = self.get_paths(x1, y1, x2, y2, panel_params, coord, ax)
                coll = PathCollection(paths, **d)
                ax.add_collection(coll)
            if last:
                x1, y1, x2, y2 = x2, y2, x1, y1
                paths = self.get_paths(x1, y1, x2, y2, panel_params, coord, ax)
                coll = PathCollection(paths, **d)
                ax.add_collection(coll)
        else:
            from matplotlib.patches import PathPatch

            d = {
                "zorder": params["zorder"],
                "rasterized": params["raster"],
                "edgecolor": data["color"].iloc[0],
                "facecolor": data["facecolor"].iloc[0],
                "linewidth": data["size"].iloc[0],
                "linestyle": data["linetype"].iloc[0],
                "joinstyle": "round",
                "capstyle": "butt",
            }

            if first:
                x1, x2 = data["x"].iloc[0:2]
                y1, y2 = data["y"].iloc[0:2]
                x1, y1, x2, y2 = (np.array([i]) for i in (x1, y1, x2, y2))
                paths = self.get_paths(x1, y1, x2, y2, panel_params, coord, ax)
                patch = PathPatch(paths[0], **d)
                ax.add_artist(patch)

            if last:
                x1, x2 = data["x"].iloc[-2:]
                y1, y2 = data["y"].iloc[-2:]
                x1, y1, x2, y2 = x2, y2, x1, y1
                x1, y1, x2, y2 = (np.array([i]) for i in (x1, y1, x2, y2))
                paths = self.get_paths(x1, y1, x2, y2, panel_params, coord, ax)
                patch = PathPatch(paths[0], **d)
                ax.add_artist(patch)

    def get_paths(
        self,
        x1: npt.ArrayLike,
        y1: npt.ArrayLike,
        x2: npt.ArrayLike,
        y2: npt.ArrayLike,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
    ) -> list[Path]:
        """
        Compute paths that create the arrow heads

        Parameters
        ----------
        x1, y1, x2, y2 : array_like
            List of points that define the tails of the arrows.
            The arrow heads will be at x1, y1. If you need them
            at x2, y2 reverse the input.

        panel_params : panel_view
            The scale information as may be required by the
            axes. At this point, that information is about
            ranges, ticks and labels. Attributes are of interest
            to the geom are::

                'panel_params.x.range'  # tuple
                'panel_params.y.range'  # tuple

        coord : coord
            Coordinate (e.g. coord_cartesian) system of the
            geom.
        ax : axes
            Axes on which to plot.

        Returns
        -------
        out : list of Path
            Paths that create arrow heads
        """
        from matplotlib.path import Path

        # The arrowhead path has 3 vertices,
        # plus a dummy vertex for the STOP code
        dummy = (0, 0)

        # codes list remains the same after initialization
        codes = [Path.MOVETO, Path.LINETO, Path.LINETO, Path.STOP]

        # We need the axes dimensions so that we can
        # compute scaling factors
        width, height = _axes_get_size_inches(ax)
        ranges = coord.range(panel_params)
        width_ = np.ptp(ranges.x)
        height_ = np.ptp(ranges.y)

        # scaling factors to prevent skewed arrowheads
        lx = self.length * width_ / width
        ly = self.length * height_ / height

        # angle in radians
        a = self.angle * np.pi / 180

        # direction of arrow head
        xdiff, ydiff = x2 - x1, y2 - y1  # type: ignore
        rotations = np.arctan2(ydiff / ly, xdiff / lx)

        # Arrow head vertices
        v1x = x1 + lx * np.cos(rotations + a)
        v1y = y1 + ly * np.sin(rotations + a)
        v2x = x1 + lx * np.cos(rotations - a)
        v2y = y1 + ly * np.sin(rotations - a)

        # create a path for each arrow head
        paths = []
        for t in zip(v1x, v1y, x1, y1, v2x, v2y):  # type: ignore
            verts = [t[:2], t[2:4], t[4:], dummy]
            paths.append(Path(verts, codes))

        return paths


def _draw_segments(data: pd.DataFrame, ax: Axes, **params: Any):
    """
    Draw independent line segments between all the
    points
    """
    from matplotlib.collections import LineCollection

    color = to_rgba(data["color"], data["alpha"])
    # All we do is line-up all the points in a group
    # into segments, all in a single list.
    # Along the way the other parameters are put in
    # sequences accordingly
    indices: list[int] = []  # for attributes of starting point of each segment
    _segments = []
    for _, df in data.groupby("group"):
        idx = df.index
        indices.extend(idx[:-1].to_list())  # One line from two points
        x = data["x"].iloc[idx]
        y = data["y"].iloc[idx]
        _segments.append(make_line_segments(x, y, ispath=True))

    segments = np.vstack(_segments)

    if color is None:
        edgecolor = color
    else:
        edgecolor = [color[i] for i in indices]

    linewidth = data.loc[indices, "size"]
    linestyle = data.loc[indices, "linetype"]

    coll = LineCollection(
        segments,
        edgecolor=edgecolor,
        linewidth=linewidth,
        linestyle=linestyle,
        zorder=params["zorder"],
        rasterized=params["raster"],
    )
    ax.add_collection(coll)


def _draw_lines(data: pd.DataFrame, ax: Axes, **params: Any):
    """
    Draw a path with the same characteristics from the
    first point to the last point
    """
    from matplotlib.lines import Line2D

    color = to_rgba(data["color"].iloc[0], data["alpha"].iloc[0])
    join_style = _get_joinstyle(data, params)
    lines = Line2D(
        data["x"],
        data["y"],
        color=color,
        linewidth=data["size"].iloc[0],
        linestyle=data["linetype"].iloc[0],
        zorder=params["zorder"],
        rasterized=params["raster"],
        **join_style,
    )
    ax.add_artist(lines)


def _get_joinstyle(
    data: pd.DataFrame, params: dict[str, Any]
) -> dict[str, Any]:
    with suppress(KeyError):
        if params["linejoin"] == "mitre":
            params["linejoin"] = "miter"

    with suppress(KeyError):
        if params["lineend"] == "square":
            params["lineend"] = "projecting"

    joinstyle = params.get("linejoin", "miter")
    capstyle = params.get("lineend", "butt")
    d = {}
    if data["linetype"].iloc[0] == "solid":
        d["solid_joinstyle"] = joinstyle
        d["solid_capstyle"] = capstyle
    elif data["linetype"].iloc[0] == "dashed":
        d["dash_joinstyle"] = joinstyle
        d["dash_capstyle"] = capstyle
    return d


def _axes_get_size_inches(ax: Axes) -> TupleFloat2:
    """
    Size of axes in inches

    Parameters
    ----------
    ax : axes
        Axes

    Returns
    -------
    out : tuple[float, float]
        (width, height) of ax in inches
    """
    fig = ax.get_figure()
    bbox = ax.get_window_extent().transformed(
        fig.dpi_scale_trans.inverted()  # pyright: ignore
    )
    return bbox.width, bbox.height



import numpy as np
import pandas as pd
import pytest

from plotnine import (
    aes,
    annotation_stripes,
    coord_flip,
    facet_wrap,
    geom_point,
    geom_vline,
    ggplot,
)

n = 9
df = pd.DataFrame({"x": np.arange(n) % 3 + 1, "y": range(n)})


def test_annotation_stripes():
    p = (
        ggplot(df)
        + annotation_stripes(fill_range="no")
        + geom_point(aes("factor(x)", "y"))
        + geom_vline(xintercept=[0.5, 1.5, 2.5, 3.5])
    )

    assert p == "annotation_stripes"


def test_annotation_stripes_faceting():
    n = len(df)

    df2 = pd.DataFrame(
        {
            "x": np.hstack([df["x"], df["x"]]),
            "y": np.hstack([df["y"], df["y"]]),
            "g": list("a" * n + "b" * n),
        }
    )

    p = (
        ggplot()
        + annotation_stripes(fill_range="no")
        + geom_point(df2, aes("factor(x)", "y"))
        + geom_vline(xintercept=[0.5, 1.5, 2.5, 3.5])
        + facet_wrap("g")
    )
    assert p == "annotation_stripes_faceting"


def test_annotation_stripes_fill_range():
    p = (
        ggplot(df)
        + annotation_stripes()
        + geom_point(aes("factor(x)", "y"))
        + geom_vline(xintercept=[0.5, 1.5, 2.5, 3.5])
    )

    assert p == "annotation_stripes_fill_range"


def test_annotation_stripes_fill_range_cycle():
    p = (
        ggplot(df)
        + annotation_stripes(fill_range="cycle")
        + geom_point(aes("factor(x)", "y"))
        + geom_vline(xintercept=[0.5, 1.5, 2.5, 3.5])
    )

    assert p == "annotation_stripes_fill_range_cycle"


def test_annotation_stripes_coord_flip():
    p = (
        ggplot(df)
        + annotation_stripes(fill_range="no")
        + geom_point(aes("factor(x)", "y"))
        + geom_vline(xintercept=[0.5, 1.5, 2.5, 3.5])
        + coord_flip()
    )

    assert p == "annotation_stripes_coord_flip"


def test_annotation_stripes_continuous_scale():
    p = (
        ggplot(df)
        + annotation_stripes()
        + geom_point(aes("x", "y"))
        + geom_vline(xintercept=[0.5, 1.5, 2.5, 3.5])
    )

    assert p == "annotation_stripes_continuous_scale"


def test_invalid_orientation():
    with pytest.raises(ValueError):
        annotation_stripes(direction="diagonal")


def test_annotation_stripes_fill_direction_extend():
    p = (
        ggplot(df)
        + annotation_stripes(
            fill=["red", "blue", "green"],
            fill_range="no",
            direction="horizontal",
            extend=(0.15, 0.85),
            alpha=0.25,
        )
        + geom_point(aes("factor(x)", "factor(y)"))
    )

    assert p == "annotation_stripes_fill_direction_extend"


def test_annotation_stripes_single_stripe():
    p = (
        ggplot(df.assign(x=10))
        + annotation_stripes(fill=["#FF0000", "#00FF00"])
        + geom_point(aes("factor(x)", "y"))
    )

    assert p == "annotation_stripes_single_stripe"



import numpy as np
import pandas as pd
import pytest

from plotnine import aes, coord_flip, facet_grid, geom_violin, ggplot
from plotnine.data import mtcars

n = 4
m = 10
df = pd.DataFrame(
    {
        "x": np.repeat([chr(65 + i) for i in range(n)], m),
        "y": (
            [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
            + [-2, 2, 3, 4, 5, 6, 7, 8, 9, 12]
            + [-7, -5, 3, 4, 5, 6, 7, 8, 12, 15]
            + [1, 2, 3, 4, 5, 8, 8, 8, 9, 10]
        ),
    }
)


class TestAesthetics:
    p = (
        ggplot(df, aes("x"))
        + geom_violin(aes(y="y"), size=2)
        + geom_violin(df[: 2 * m], aes(y="y+25", fill="x"), size=2)
        + geom_violin(df[2 * m :], aes(y="y+25", color="x"), size=2)
        + geom_violin(df[2 * m :], aes(y="y+50", linetype="x"), size=2)
    )

    def test_aesthetics(self):
        assert self.p == "aesthetics"

    def test_aesthetics_coordflip(self):
        assert self.p + coord_flip() == "aesthetics+coord_flip"


def test_scale():
    p = (
        ggplot(df, aes("x"))
        +
        # Red should envelop blue
        geom_violin(aes(y="y"), scale="width", color="red", fill="red", size=2)
        + geom_violin(
            aes(y="y"), scale="area", color="blue", fill="blue", size=2
        )
        + geom_violin(
            df[:36], aes(y="y+25"), scale="count", color="green", size=2
        )
        +
        # Yellow should envelop green
        geom_violin(
            aes(y="y+25"), scale="count", color="yellow", fill="yellow", size=2
        )
        + geom_violin(
            df[:36],
            aes(y="y+25"),
            scale="count",
            color="green",
            fill="green",
            size=2,
        )
    )
    assert p == "scale"


def test_quantiles_width_dodge():
    p = (
        ggplot(df, aes("x"))
        + geom_violin(aes(y="y"), draw_quantiles=[0.25, 0.75], size=2)
        + geom_violin(aes(y="y+25"), color="green", width=0.5, size=2)
        + geom_violin(aes(y="y+50", fill="factor(y%2)"), size=2)
    )
    assert p == "quantiles_width_dodge"


def test_quantiles_input_checks():
    with pytest.raises(ValueError):
        geom_violin(aes("x", "y"), draw_quantiles=True)
    with pytest.raises(TypeError):
        geom_violin(aes("x", "y"), draw_quantiles=["A", 0.25])
    with pytest.raises(ValueError):
        geom_violin(aes("x", "y"), draw_quantiles=[0.25, 1.25])
    with pytest.raises(ValueError):
        geom_violin(aes("x", "y"), draw_quantiles=[0.0])
    with pytest.raises(ValueError):
        geom_violin(aes("x", "y"), draw_quantiles=[1.0])
    g = geom_violin(aes("x", "y"), draw_quantiles=np.array([0.25, 0.25]))
    assert isinstance(g.params["draw_quantiles"], np.ndarray)
    g = geom_violin(aes("x", "y"), draw_quantiles=0.5)
    assert isinstance(g.params["draw_quantiles"], np.ndarray)


def test_no_trim():
    p = ggplot(df, aes("x")) + geom_violin(aes(y="y"), trim=False, size=2)
    assert p == "no_trim"


def test_style():
    p = (
        ggplot(df, aes("x"))
        + geom_violin(aes(y="y"), style="right")
        + geom_violin(aes(y="y+25"), style="left")
    )
    assert p == "style"


def test_style_alternating():
    p = (
        ggplot(df, aes("x"))
        + geom_violin(aes(y="y"), style="right-left", fill="green")
        + geom_violin(aes(y="y+25"), style="left-right", fill="yellow")
    )
    assert p == "style_alternating"


def test_style_input_checks():
    with pytest.raises(ValueError):
        geom_violin(aes("x", "y"), style=True)
    with pytest.raises(ValueError):
        geom_violin(aes("x", "y"), style=1)
    with pytest.raises(ValueError):
        geom_violin(aes("x", "y"), style="up")


def test_scales_free():
    # GH655
    p = (
        ggplot(mtcars, aes(x="factor(cyl)", y="mpg"))
        + geom_violin()
        + facet_grid('am ~ "column"', scales="free")
    )
    assert p == "scales_free"



from __future__ import annotations

import typing

import numpy as np

from ..doctools import document
from ..scales.scale_shape import FILLED_SHAPES
from ..utils import SIZE_FACTOR, to_rgba
from .geom import geom

if typing.TYPE_CHECKING:
    from typing import Any

    import pandas as pd

    from plotnine.iapi import panel_view
    from plotnine.typing import Axes, Coord, DrawingArea, Layer


@document
class geom_point(geom):
    """
    Plot points (Scatter plot)

    {usage}

    Parameters
    ----------
    {common_parameters}
    """

    DEFAULT_AES = {
        "alpha": 1,
        "color": "black",
        "fill": None,
        "shape": "o",
        "size": 1.5,
        "stroke": 0.5,
    }
    REQUIRED_AES = {"x", "y"}
    NON_MISSING_AES = {"color", "shape", "size"}
    DEFAULT_PARAMS = {
        "stat": "identity",
        "position": "identity",
        "na_rm": False,
    }

    def draw_panel(
        self,
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        """
        Plot all groups
        """
        self.draw_group(data, panel_params, coord, ax, **params)

    @staticmethod
    def draw_group(
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        data = coord.transform(data, panel_params)
        units = "shape"
        for _, udata in data.groupby(units, dropna=False):
            udata.reset_index(inplace=True, drop=True)
            geom_point.draw_unit(udata, panel_params, coord, ax, **params)

    @staticmethod
    def draw_unit(
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        # Our size is in 'points' while scatter wants
        # 'points^2'. The stroke is outside. And pi
        # gives a large enough scaling factor
        # All other sizes for which the MPL units should
        # be in points must scaled using sqrt(pi)
        size = ((data["size"] + data["stroke"]) ** 2) * np.pi
        stroke = data["stroke"] * SIZE_FACTOR
        color = to_rgba(data["color"], data["alpha"])
        shape = data.loc[0, "shape"]

        # It is common to forget that scatter points are
        # filled and slip-up by manually assigning to the
        # color instead of the fill. We forgive.
        if shape in FILLED_SHAPES:
            if all(c is None for c in data["fill"]):
                fill = color
            else:
                fill = to_rgba(data["fill"], data["alpha"])
        else:
            # Assume unfilled
            fill = color
            color = None

        ax.scatter(
            x=data["x"],
            y=data["y"],
            s=size,
            facecolor=fill,
            edgecolor=color,
            linewidth=stroke,
            marker=shape,
            zorder=params["zorder"],
            rasterized=params["raster"],
        )

    @staticmethod
    def draw_legend(
        data: pd.Series[Any], da: DrawingArea, lyr: Layer
    ) -> DrawingArea:
        """
        Draw a point in the box

        Parameters
        ----------
        data : Series
            Data Row
        da : DrawingArea
            Canvas
        lyr : layer
            Layer

        Returns
        -------
        out : DrawingArea
        """
        from matplotlib.lines import Line2D

        if data["fill"] is None:
            data["fill"] = data["color"]

        size = (data["size"] + data["stroke"]) * SIZE_FACTOR
        stroke = data["stroke"] * SIZE_FACTOR
        fill = to_rgba(data["fill"], data["alpha"])
        color = to_rgba(data["color"], data["alpha"])

        key = Line2D(
            [0.5 * da.width],
            [0.5 * da.height],
            marker=data["shape"],
            markersize=size,
            markerfacecolor=fill,
            markeredgecolor=color,
            markeredgewidth=stroke,
        )
        da.add_artist(key)
        return da



from __future__ import annotations

import itertools
import typing
from contextlib import suppress
from typing import List
from warnings import warn

import numpy as np
import pandas.api.types as pdtypes

from ..exceptions import PlotnineError, PlotnineWarning
from ..mapping.aes import aes_to_scale
from ..utils import Registry, array_kind
from .scale import scale

if typing.TYPE_CHECKING:
    import pandas as pd

    from plotnine.typing import (
        Scale,
        ScaledAestheticsName,
    )


_TPL_DUPLICATE_SCALE = """\
Scale for '{0}' is already present.
Adding another scale for '{0}',
which will replace the existing scale.
"""


class Scales(List[scale]):
    """
    List of scales

    This class has methods the simplify the handling of
    the ggplot object scales
    """

    def append(self, sc: Scale):
        """
        Add / Update scale

        Removes any previous scales that cover the same aesthetics
        """
        ae = sc.aesthetics[0]
        cover_ae = self.find(ae)
        if any(cover_ae):
            warn(_TPL_DUPLICATE_SCALE.format(ae), PlotnineWarning)
            idx = cover_ae.index(True)
            self.pop(idx)
        # super() does not work well with reloads
        list.append(self, sc)

    def find(self, aesthetic: ScaledAestheticsName) -> list[bool]:
        """
        Find scales for given aesthetic

        Returns a list[bool] each scale if it covers the aesthetic
        """
        return [aesthetic in s.aesthetics for s in self]

    def input(self):
        """
        Return a list of all the aesthetics covered by the scales
        """
        lst = [s.aesthetics for s in self]
        return list(itertools.chain(*lst))

    def get_scales(self, aesthetic: ScaledAestheticsName) -> Scale | None:
        """
        Return the scale for the aesthetic or None if there isn't one

        These are the scales specified by the user e.g
            `ggplot() + scale_x_continuous()`
        or those added by default during the plot building
        process
        """
        bool_lst = self.find(aesthetic)
        try:
            idx = bool_lst.index(True)
            return self[idx]
        except ValueError:
            return None

    @property
    def x(self) -> Scale | None:
        """
        Return x scale
        """
        return self.get_scales("x")

    @property
    def y(self) -> Scale | None:
        """
        Return y scale
        """
        return self.get_scales("y")

    def non_position_scales(self) -> Scales:
        """
        Return a list of any non-position scales
        """
        l = [
            s
            for s in self
            if "x" not in s.aesthetics and "y" not in s.aesthetics
        ]
        return Scales(l)

    def position_scales(self) -> Scales:
        """
        Return a list of the position scales that are present
        """
        l = [s for s in self if ("x" in s.aesthetics) or ("y" in s.aesthetics)]
        return Scales(l)

    def train(self, data, vars, idx):
        """
        Train the scales on the data.

        The scales should be for the same aesthetic
        e.g. x scales, y scales, color scales, ...

        Parameters
        ----------
        data : dataframe
            data to use for training
        vars : list | tuple
            columns in data to use for training.
            These should be all the aesthetics of
            a scale type that are present in the
            data. e.g x, xmin, xmax
        idx : array-like
            indices that map the data points to the
            scales. These start at 1, so subtract 1 to
            get the true index into the scales array
        """
        idx = np.asarray(idx)
        for col in vars:
            for i, sc in enumerate(self, start=1):
                bool_idx = i == idx
                sc.train(data.loc[bool_idx, col])

    def map(self, data, vars, idx):
        """
        Map the data onto the scales

        The scales should be for the same aesthetic
        e.g. x scales, y scales, color scales, ...

        Parameters
        ----------
        data : dataframe
            data with columns to map
            This is modified inplace
        vars : list | tuple
            columns to map
        idx : array-like
            indices that link the data points to the
            scales. These start at 1, so subtract 1 to
            get the true index into the scales array
        """
        idx = np.asarray(idx)
        # discrete scales change the dtype
        # from category to int. Use a new dataframe
        # to collect these results.
        # Using `type` preserves the subclass of pd.DataFrame
        df = type(data)(index=data.index)
        discrete_cols = []

        # Loop through each variable, mapping across each scale,
        # then joining back into the copy of the data
        for col in vars:
            use_df = array_kind.discrete(data[col])
            if use_df:
                discrete_cols.append(col)
            for i, sc in enumerate(self, start=1):
                bool_idx = i == idx
                results = sc.map(data.loc[bool_idx, col])
                if use_df:
                    df.loc[bool_idx, col] = results
                else:
                    data.loc[bool_idx, col] = results

        for col in discrete_cols:
            data[col] = df[col]

    def reset(self):
        """
        Reset all the scales
        """
        for sc in self:
            sc.reset()

    def train_df(self, df: pd.DataFrame, drop: bool = False):
        """
        Train scales from a dataframe
        """
        if (len(df) == 0) or (len(self) == 0):
            return

        # Each scale trains the columns it understands
        for sc in self:
            sc.train_df(df)

    def map_df(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Map values from a dataframe.

        Returns dataframe
        """
        if (len(df) == 0) or (len(self) == 0):
            return df

        # Each scale maps the columns it understands
        for sc in self:
            df = sc.map_df(df)
        return df

    def transform_df(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Transform values in a dataframe.

        Returns dataframe
        """
        if (len(df) == 0) or (len(self) == 0):
            return df

        # Each scale transforms the columns it understands
        for sc in self:
            df = sc.transform_df(df)
        return df

    def add_defaults(self, data, aesthetics):
        """
        Add default scales for the aesthetics if there is none

        Scales are added only if the aesthetic is mapped to
        a column in the dataframe. This function may have to be
        called separately after evaluating the aesthetics.
        """
        if not aesthetics:
            return

        # aesthetics with scales
        aws = set()
        if self:
            for s in (set(sc.aesthetics) for sc in self):
                aws.update(s)

        # aesthetics that do not have scales present
        # We preserve the order of the aesthetics
        new_aesthetics = [x for x in aesthetics.keys() if x not in aws]
        if not new_aesthetics:
            return

        # If a new aesthetic corresponds to a column in the data
        # frame, find a default scale for the type of data in that
        # column
        seen = set()
        for ae in new_aesthetics:
            col = aesthetics[ae]
            if col not in data:
                col = ae
            scale_var = aes_to_scale(ae)

            if self.get_scales(scale_var):
                continue

            seen.add(scale_var)
            try:
                sc = make_scale(scale_var, data[col])
            except PlotnineError:
                # Skip aesthetics with no scales (e.g. group, order, etc)
                continue
            self.append(sc)

    def add_missing(self, aesthetics):
        """
        Add missing but required scales.

        Parameters
        ----------
        aesthetics : list | tuple
            Aesthetic names. Typically, ('x', 'y').
        """
        # Keep only aesthetics that don't have scales
        aesthetics = set(aesthetics) - set(self.input())

        for ae in aesthetics:
            scale_name = f"scale_{ae}_continuous"
            scale_f = Registry[scale_name]
            self.append(scale_f())


def scale_type(series):
    """
    Get a suitable scale for the series
    """
    if array_kind.continuous(series):
        stype = "continuous"
    elif array_kind.ordinal(series):
        stype = "ordinal"
    elif array_kind.discrete(series):
        stype = "discrete"
    elif array_kind.datetime(series):
        stype = "datetime"
    elif array_kind.timedelta(series):
        stype = "timedelta"
    else:
        msg = (
            "Don't know how to automatically pick scale for "
            "object of type {}. Defaulting to 'continuous'"
        )
        warn(msg.format(series.dtype), PlotnineWarning)
        stype = "continuous"
    return stype


def make_scale(ae, series, *args, **kwargs):
    """
    Return a proper scale object for the series

    The scale is for the aesthetic ae, and args & kwargs
    are passed on to the scale creating class
    """
    if pdtypes.is_float_dtype(series) and np.isinf(series).all():
        raise PlotnineError("Cannot create scale for infinite data")

    stype = scale_type(series)

    # filter parameters by scale type
    if stype in ("discrete", "ordinal"):
        with suppress(KeyError):
            del kwargs["trans"]

    scale_name = f"scale_{ae}_{stype}"
    scale_klass = Registry[scale_name]
    return scale_klass(*args, **kwargs)



"""
Plotting objects
"""

from .annotate import annotate
from .annotation_logticks import annotation_logticks
from .annotation_stripes import annotation_stripes
from .geom_abline import geom_abline
from .geom_area import geom_area
from .geom_bar import geom_bar
from .geom_bin_2d import geom_bin2d, geom_bin_2d
from .geom_blank import geom_blank
from .geom_boxplot import geom_boxplot
from .geom_col import geom_col
from .geom_count import geom_count
from .geom_crossbar import geom_crossbar
from .geom_density import geom_density
from .geom_density_2d import geom_density_2d
from .geom_dotplot import geom_dotplot
from .geom_errorbar import geom_errorbar
from .geom_errorbarh import geom_errorbarh
from .geom_freqpoly import geom_freqpoly
from .geom_histogram import geom_histogram
from .geom_hline import geom_hline
from .geom_jitter import geom_jitter
from .geom_label import geom_label
from .geom_line import geom_line
from .geom_linerange import geom_linerange
from .geom_map import geom_map
from .geom_path import arrow, geom_path
from .geom_point import geom_point
from .geom_pointdensity import geom_pointdensity
from .geom_pointrange import geom_pointrange
from .geom_polygon import geom_polygon
from .geom_qq import geom_qq
from .geom_qq_line import geom_qq_line
from .geom_quantile import geom_quantile
from .geom_raster import geom_raster
from .geom_rect import geom_rect
from .geom_ribbon import geom_ribbon
from .geom_rug import geom_rug
from .geom_segment import geom_segment
from .geom_sina import geom_sina
from .geom_smooth import geom_smooth
from .geom_spoke import geom_spoke
from .geom_step import geom_step
from .geom_text import geom_text
from .geom_tile import geom_tile
from .geom_violin import geom_violin
from .geom_vline import geom_vline

__all__ = [
    "annotate",
    "annotation_logticks",
    "annotation_stripes",
    "geom_abline",
    "geom_area",
    "geom_bar",
    "geom_bin_2d",
    "geom_bin2d",
    "geom_blank",
    "geom_boxplot",
    "geom_col",
    "geom_count",
    "geom_crossbar",
    "geom_density",
    "geom_density_2d",
    "geom_dotplot",
    "geom_errorbar",
    "geom_errorbarh",
    "geom_freqpoly",
    "geom_histogram",
    "geom_hline",
    "geom_jitter",
    "geom_label",
    "geom_line",
    "geom_linerange",
    "geom_map",
    "arrow",
    "geom_path",
    "geom_point",
    "geom_pointdensity",
    "geom_pointrange",
    "geom_quantile",
    "geom_qq",
    "geom_qq_line",
    "geom_polygon",
    "geom_raster",
    "geom_rect",
    "geom_ribbon",
    "geom_rug",
    "geom_segment",
    "geom_sina",
    "geom_smooth",
    "geom_spoke",
    "geom_step",
    "geom_text",
    "geom_tile",
    "geom_violin",
    "geom_vline",
]



from __future__ import annotations

import typing

from ..doctools import document
from ..utils import to_rgba
from .geom_text import geom_text

if typing.TYPE_CHECKING:
    from typing import Any

    import pandas as pd

    from plotnine.typing import DrawingArea, Layer


_aes = geom_text.DEFAULT_AES.copy()
_aes["fill"] = "white"

_params = geom_text.DEFAULT_PARAMS.copy()
_params.update(
    {
        # boxstyle is one of
        #   cirle, larrow, rarrow, round, round4,
        #   roundtooth, sawtooth, square
        #
        # Documentation at matplotlib.patches.BoxStyle
        "boxstyle": "round",
        "boxcolor": None,
        "label_padding": 0.25,
        "label_r": 0.25,
        "label_size": 0.7,
        "tooth_size": None,
    }
)


@document
class geom_label(geom_text):
    """
    Textual annotations with a background

    {usage}

    Parameters
    ----------
    {common_parameters}
    boxstyle : str, optional (default: round)
        Options are::

            'circle'
            'darrow'
            'larrow'
            'rarrow'
            'round '
            'round4'
            'roundtooth'
            'sawtooth'
            'square'
    boxcolor: None, str or rgba tuple (default: None)
        Color of box around the text. If None, the color is
        the same as the text.
    label_padding : float, optional (default: 0.25)
        Amount of padding
    label_r : float, optional (default: 0.25)
        Rounding radius of corners.
    label_size : float, optional (default: 0.7)
        Linewidth of the label boundary.
    tooth_size : float, optional (default: None)
        Size of the ``roundtooth`` or ``sawtooth`` if they
        are the chosen *boxstyle*. The default depends
        on Matplotlib

    See Also
    --------
    plotnine.geoms.geom_text : For documentation of the
        parameters. :class:`matplotlib.patches.BoxStyle` for the
        parameters that affect the boxstyle.
    """

    DEFAULT_AES = _aes
    DEFAULT_PARAMS = _params

    @staticmethod
    def draw_legend(
        data: pd.Series[Any], da: DrawingArea, lyr: Layer
    ) -> DrawingArea:
        """
        Draw letter 'a' in the box

        Parameters
        ----------
        data : Series
            Data Row
        da : DrawingArea
            Canvas
        lyr : layer
            Layer

        Returns
        -------
        out : DrawingArea
        """
        from matplotlib.patches import Rectangle

        fill = to_rgba(data["fill"], data["alpha"])

        if data["fill"]:
            rect = Rectangle(
                (0, 0),
                width=da.width,
                height=da.height,
                linewidth=0,
                facecolor=fill,
                capstyle="projecting",
            )
            da.add_artist(rect)
        return geom_text.draw_legend(data, da, lyr)



from ..options import get_option
from .elements import element_rect, element_text
from .theme import theme


class theme_seaborn(theme):
    """
    Theme for seaborn.

    Credit to Michael Waskom's seaborn:

        - http://stanford.edu/~mwaskom/software/seaborn
        - https://github.com/mwaskom/seaborn

    Parameters
    ----------
    style: str in ``['whitegrid', 'darkgrid', 'nogrid', 'ticks']``
        Style of axis background.
    context: str in ``['notebook', 'talk', 'paper', 'poster']``
        Intended context for resulting figures.
    font : str
        Font family, see matplotlib font manager.
    font_scale : float, optional
        Separate scaling factor to independently scale the
        size of the font elements.
    """

    def __init__(
        self,
        style="darkgrid",
        context="notebook",
        font="sans-serif",
        font_scale=1,
    ):
        from .seaborn_rcmod import set_theme

        rcparams = set_theme(
            context=context, style=style, font=font, font_scale=font_scale
        )
        base_size = rcparams["font.size"]
        half_line = base_size / 2
        line_margin = half_line * 0.8 / 2
        m = get_option("base_margin")

        super().__init__(
            aspect_ratio=get_option("aspect_ratio"),
            dpi=get_option("dpi"),
            figure_size=get_option("figure_size"),
            text=element_text(size=base_size, rotation=0),
            axis_text=element_text(
                size=base_size * 0.8,
                margin={
                    "t": line_margin,
                    "b": line_margin,
                    "l": line_margin,
                    "r": line_margin,
                    "units": "pt",
                },
            ),
            axis_title_x=element_text(
                va="bottom", ha="center", margin={"t": m, "units": "fig"}
            ),
            axis_title_y=element_text(
                angle=90,
                va="center",
                ha="left",
                margin={"r": m, "units": "fig"},
            ),
            legend_box="auto",
            legend_box_just="auto",
            legend_box_margin=0,
            legend_box_spacing=m * 3,  # figure units
            legend_direction="auto",
            legend_entry_spacing_x=5,
            legend_entry_spacing_y=2,
            legend_key_size=base_size * 0.8 * 1.8,
            legend_margin=0,
            legend_position="right",
            legend_spacing=10,  # points
            legend_text=element_text(
                va="baseline",
                ha="left",
                margin={"t": 3, "b": 3, "l": 3, "r": 3, "units": "pt"},
            ),
            legend_title=element_text(
                ha="left",
                margin={
                    "t": m,
                    "b": m,
                    "l": m,
                    "r": m,
                    "units": "fig",
                },
            ),
            legend_title_align="auto",
            panel_spacing=m,
            plot_caption=element_text(
                size=base_size * 0.8,
                ha="right",
                va="bottom",
                ma="left",
                margin={"t": m, "units": "fig"},
            ),
            plot_margin=m,
            plot_subtitle=element_text(
                size=base_size * 1,
                ha="left",
                va="top",
                ma="left",
                margin={"b": m, "units": "fig"},
            ),
            plot_title=element_text(
                size=base_size * 1.2,
                ha="left",
                va="top",
                ma="left",
                margin={"b": m, "units": "fig"},
            ),
            strip_align=0,
            strip_background=element_rect(color="None", fill="#D1CDDF"),
            strip_text=element_text(
                size=base_size * 0.8,
                linespacing=1.0,
                margin={
                    "t": 1 / 3,
                    "b": 1 / 3,
                    "l": 1 / 3,
                    "r": 1 / 3,
                    "units": "lines",
                },
            ),
            strip_text_y=element_text(rotation=-90),
            complete=True,
        )

        self._rcParams.update(rcparams)



from __future__ import annotations

import typing
from warnings import warn

import numpy as np
import pandas as pd

from ..doctools import document
from ..exceptions import PlotnineWarning
from ..positions import position_dodge2
from ..positions.position import position
from ..utils import (
    SIZE_FACTOR,
    copy_missing_columns,
    resolution,
    to_rgba,
)
from .geom import geom
from .geom_crossbar import geom_crossbar
from .geom_point import geom_point
from .geom_segment import geom_segment

if typing.TYPE_CHECKING:
    from typing import Any

    import numpy.typing as npt

    from plotnine.iapi import panel_view
    from plotnine.typing import (
        Aes,
        Axes,
        Coord,
        DataLike,
        DrawingArea,
        Layer,
    )


@document
class geom_boxplot(geom):
    """
    Box and whiskers plot

    {usage}

    Parameters
    ----------
    {common_parameters}
    width : float, optional (default None)
        Box width. If :py:`None`, the width is set to
        `90%` of the resolution of the data. Note that if the stat
        has a width parameter, that takes precedence over this one.
    outlier_alpha : float, optional (default: 1)
        Transparency of the outlier points.
    outlier_color : str or tuple, optional (default: None)
        Color of the outlier points.
    outlier_shape : str, optional (default: o)
        Shape of the outlier points. An empty string hides the outliers.
    outlier_size : float, optional (default: 1.5)
        Size of the outlier points.
    outlier_stroke : float, optional (default: 0.5)
        Stroke-size of the outlier points.
    notch : bool, optional (default: False)
        Whether the boxes should have a notch.
    varwidth : bool, optional (default: False)
        If :py:`True`, boxes are drawn with widths proportional to
        the square-roots of the number of observations in the
        groups.
    notchwidth : float, optional (default: 0.5)
        Width of notch relative to the body width.
    fatten : float, optional (default: 2)
        A multiplicative factor used to increase the size of the
        middle bar across the box.
    """

    DEFAULT_AES = {
        "alpha": 1,
        "color": "#333333",
        "fill": "white",
        "linetype": "solid",
        "shape": "o",
        "size": 0.5,
        "weight": 1,
    }
    REQUIRED_AES = {"x", "lower", "upper", "middle", "ymin", "ymax"}
    DEFAULT_PARAMS = {
        "stat": "boxplot",
        "position": "dodge2",
        "na_rm": False,
        "width": None,
        "outlier_alpha": 1,
        "outlier_color": None,
        "outlier_shape": "o",
        "outlier_size": 1.5,
        "outlier_stroke": 0.5,
        "notch": False,
        "varwidth": False,
        "notchwidth": 0.5,
        "fatten": 2,
    }

    def __init__(
        self,
        mapping: Aes | None = None,
        data: DataLike | None = None,
        **kwargs: Any,
    ):
        _position = kwargs.get("position", self.DEFAULT_PARAMS["position"])
        varwidth = kwargs.get("varwidth", self.DEFAULT_PARAMS["varwidth"])

        # varwidth = True is not compatible with preserve='total'
        if varwidth:
            if isinstance(_position, str):
                kwargs["position"] = position_dodge2(preserve="single")
            elif isinstance(_position, position):
                if _position.params["preserve"] == "total":
                    warn(
                        "Cannot preserve total widths when varwidth=True",
                        PlotnineWarning,
                    )
                    _position.params["preserve"] = "single"

        super().__init__(mapping, data, **kwargs)

    def setup_data(self, data: pd.DataFrame) -> pd.DataFrame:
        if "width" not in data:
            width = self.params.get("width", None)
            if width is not None:
                data["width"] = width
            else:
                data["width"] = resolution(data["x"], False) * 0.9

        if "outliers" not in data:
            data["outliers"] = [[] for i in range(len(data))]

        # min and max outlier values
        omin = [
            np.min(lst) if len(lst) else +np.inf for lst in data["outliers"]
        ]
        omax = [
            np.max(lst) if len(lst) else -np.inf for lst in data["outliers"]
        ]

        data["ymin_final"] = np.min(
            np.column_stack([data["ymin"], omin]), axis=1
        )
        data["ymax_final"] = np.max(
            np.column_stack([data["ymax"], omax]), axis=1
        )

        # if varwidth not requested or not available, don't use it
        if (
            "varwidth" not in self.params
            or not self.params["varwidth"]
            or "relvarwidth" not in data
        ):
            data["xmin"] = data["x"] - data["width"] / 2
            data["xmax"] = data["x"] + data["width"] / 2
        else:
            # make relvarwidth relative to the size of the
            # largest group
            data["relvarwidth"] /= data["relvarwidth"].max()
            data["xmin"] = data["x"] - data["relvarwidth"] * data["width"] / 2
            data["xmax"] = data["x"] + data["relvarwidth"] * data["width"] / 2
            del data["relvarwidth"]

        del data["width"]

        return data

    @staticmethod
    def draw_group(
        data: pd.DataFrame,
        panel_params: panel_view,
        coord: Coord,
        ax: Axes,
        **params: Any,
    ):
        def flat(*args: pd.Series[Any]) -> npt.NDArray[Any]:
            """Flatten list-likes"""
            return np.hstack(args)

        common_columns = [
            "color",
            "size",
            "linetype",
            "fill",
            "group",
            "alpha",
            "shape",
        ]
        # whiskers
        whiskers = pd.DataFrame(
            {
                "x": flat(data["x"], data["x"]),
                "y": flat(data["upper"], data["lower"]),
                "yend": flat(data["ymax"], data["ymin"]),
                "alpha": 1,
            }
        )
        whiskers["xend"] = whiskers["x"]
        copy_missing_columns(whiskers, data[common_columns])

        # box
        box_columns = ["xmin", "xmax", "lower", "middle", "upper"]
        box = data[common_columns + box_columns].copy()
        box.rename(
            columns={"lower": "ymin", "middle": "y", "upper": "ymax"},
            inplace=True,
        )

        # notch
        if params["notch"]:
            box["ynotchlower"] = data["notchlower"]
            box["ynotchupper"] = data["notchupper"]

        # outliers
        num_outliers = len(data["outliers"].iloc[0])
        if num_outliers:

            def outlier_value(param: str) -> Any:
                oparam = f"outlier_{param}"
                if params[oparam] is not None:
                    return params[oparam]
                return data[param].iloc[0]

            outliers = pd.DataFrame(
                {
                    "y": data["outliers"].iloc[0],
                    "x": np.repeat(data["x"].iloc[0], num_outliers),
                    "fill": [None] * num_outliers,
                }
            )
            outliers["alpha"] = outlier_value("alpha")
            outliers["color"] = outlier_value("color")
            outliers["shape"] = outlier_value("shape")
            outliers["size"] = outlier_value("size")
            outliers["stroke"] = outlier_value("stroke")
            geom_point.draw_group(outliers, panel_params, coord, ax, **params)

        # plot
        geom_segment.draw_group(whiskers, panel_params, coord, ax, **params)
        geom_crossbar.draw_group(box, panel_params, coord, ax, **params)

    @staticmethod
    def draw_legend(
        data: pd.Series[Any], da: DrawingArea, lyr: Layer
    ) -> DrawingArea:
        """
        Draw a rectangle in the box

        Parameters
        ----------
        data : Series
            Data Row
        da : DrawingArea
            Canvas
        lyr : layer
            Layer

        Returns
        -------
        out : DrawingArea
        """
        from matplotlib.lines import Line2D
        from matplotlib.patches import Rectangle

        # box
        facecolor = to_rgba(data["fill"], data["alpha"])
        if facecolor is None:
            facecolor = "none"

        kwargs = {"linestyle": data["linetype"], "linewidth": data["size"]}

        box = Rectangle(
            (da.width * 0.125, da.height * 0.25),
            width=da.width * 0.75,
            height=da.height * 0.5,
            facecolor=facecolor,
            edgecolor=data["color"],
            capstyle="projecting",
            antialiased=False,
            **kwargs,
        )
        da.add_artist(box)

        kwargs["solid_capstyle"] = "butt"
        kwargs["color"] = data["color"]
        kwargs["linewidth"] *= SIZE_FACTOR

        # middle strike through
        strike = Line2D(
            [da.width * 0.125, da.width * 0.875],
            [da.height * 0.5, da.height * 0.5],
            **kwargs,
        )
        da.add_artist(strike)

        # whiskers
        top = Line2D(
            [da.width * 0.5, da.width * 0.5],
            [da.height * 0.75, da.height * 0.9],
            **kwargs,
        )
        da.add_artist(top)

        bottom = Line2D(
            [da.width * 0.5, da.width * 0.5],
            [da.height * 0.25, da.height * 0.1],
            **kwargs,
        )
        da.add_artist(bottom)
        return da



import numpy as np

from plotnine.scales import scale_x_continuous, scale_x_discrete
from plotnine.stats.binning import (
    _adjust_breaks,
    breaks_from_bins,
    breaks_from_binwidth,
    fuzzybreaks,
)


def test_breaks_from_bins():
    n = 10
    x = list(range(n))
    limits = min(x), max(x)
    breaks = breaks_from_bins(limits, n)
    breaks2 = breaks_from_bins(limits, n, center=limits[0])

    # correct no. of breaks
    assert len(breaks) - 1 == n
    # breaks include the limits
    assert min(breaks) <= limits[0]
    assert max(breaks) >= limits[1]
    # limits are at mid-points of the bins in which they fall
    assert (breaks[1] + breaks[0]) / 2 == limits[0]
    assert (breaks[-1] + breaks[-2]) / 2 == limits[1]
    assert list(breaks) == list(breaks2)

    breaks = breaks_from_bins(limits, n, boundary=limits[0])
    assert list(breaks) == x


def test_breaks_from_binwidth():
    n = 10
    x = list(range(n))
    limits = min(x), max(x)
    breaks = breaks_from_binwidth(limits, binwidth=1)
    breaks2 = breaks_from_binwidth(limits, binwidth=1, center=limits[0])

    # correct no. of breaks
    assert len(breaks) - 1 == n
    # breaks include the limits
    assert min(breaks) <= limits[0]
    assert max(breaks) >= limits[1]
    # limits are at mid-points of the bins in which they fall
    assert (breaks[1] + breaks[0]) / 2 == limits[0]
    assert (breaks[-1] + breaks[-2]) / 2 == limits[1]
    assert list(breaks) == list(breaks2)


def test_fuzzybreaks():
    n = 10
    x = list(range(n))
    limits = min(x), max(x)
    # continuous scale
    cscale = scale_x_continuous(limits=limits)
    cscale.train(x)

    breaks = fuzzybreaks(cscale, bins=n - 1, right=True)
    assert breaks[0] <= limits[0]
    assert breaks[-1] >= limits[1]
    assert all(_x < b for _x, b in zip(x[1:-1], breaks[1:-1]))

    breaks = fuzzybreaks(cscale, bins=n - 1, right=False)
    assert breaks[0] <= limits[0]
    assert breaks[-1] >= limits[1]
    assert all(b < _x for _x, b in zip(x[1:-1], breaks[1:-1]))

    breaks = fuzzybreaks(cscale, binwidth=1, right=True)
    assert breaks[0] <= limits[0]
    assert breaks[-1] >= limits[1]
    assert all(_x < b for _x, b in zip(x[1:-1], breaks[1:-1]))

    # discrete scale
    x = list(range(1, n + 1))  # items are "labelled" 1 to n
    dscale = scale_x_discrete(limits=x)
    breaks = fuzzybreaks(dscale)
    # The breaks create bins centered on the limits
    assert list((breaks[:-1] + breaks[1:]) / 2) == x


def test_adjust_breaks_right():
    def _test(a, b):
        assert b[0] <= a[0]
        assert all(a[1:] <= b[1:])

    # All positive
    a = np.linspace(1, 2, 11)
    b = _adjust_breaks(a, right=True)
    _test(a, b)

    # zero on the Left
    a = np.linspace(0, 1, 11)
    b = _adjust_breaks(a, right=True)
    _test(a, b)

    # zero on the right
    a = np.linspace(-1, 0, 11)
    b = _adjust_breaks(a, right=True)
    _test(a, b)

    # All negative
    a = np.linspace(-2, -1, 11)
    b = _adjust_breaks(a, right=True)
    _test(a, b)


def test_adjust_breaks_right_False():
    def _test(a, b):
        assert a[-1] <= b[-1]
        assert all(b[:-1] <= a[:-1])

    # All positive
    a = np.linspace(1, 2, 11)
    b = _adjust_breaks(a, right=False)
    _test(a, b)

    # zero on the Left
    a = np.linspace(0, 1, 11)
    b = _adjust_breaks(a, right=False)
    _test(a, b)

    # zero on the right
    a = np.linspace(-1, 0, 11)
    b = _adjust_breaks(a, right=False)
    _test(a, b)

    # All negative
    a = np.linspace(-2, -1, 11)
    b = _adjust_breaks(a, right=False)
    _test(a, b)



import numpy as np
import pandas as pd
import pytest

from plotnine import qplot
from plotnine.exceptions import PlotnineWarning


def test_scalars():
    p = qplot(x=2, y=3)
    assert p == "scalars"


def test_arrays():
    p = qplot(x=np.arange(5), y=np.arange(5))
    assert p == "arrays"


def test_string_arrays():
    p = qplot(x="np.arange(5)", y="np.arange(5)")
    assert p == "string-arrays"


def test_range():
    p = qplot(x=range(5), y=range(5))

    assert p == "range"


def test_onlyx():
    p = qplot(x="np.arange(5)")
    with pytest.warns(PlotnineWarning):
        assert p == "onlyx"


def test_onlyy():
    p = qplot(y=np.arange(5))

    assert p == "onlyy"


def test_sample():
    p = qplot(sample="np.arange(5)")
    assert p == "sample"


def test_xlim():
    p = qplot(x="np.arange(5)", y="np.arange(5)", xlim=(-10, 10))
    assert p == "xlim"


def test_ylim():
    p = qplot(x="np.arange(5)", y="np.arange(5)", ylim=(-10, 10))
    assert p == "ylim"


def test_multiple_geoms():
    n = 3
    m = 10
    # n stairs of points, each m points high
    df = pd.DataFrame(
        {"x": np.repeat(range(n), m), "y": np.linspace(0, n, n * m)}
    )
    p = qplot("factor(x)", "y", data=df, geom=("boxplot", "point"))
    assert p == "multiple_geoms"


def test_series_labelling():
    df = pd.DataFrame(
        {
            "x_axis_label": [1, 2, 3],
            "y_axis_label": [1, 2, 3],
            "color_label": ["a", "b", "c"],
        }
    )
    p = qplot(df.x_axis_label, df.y_axis_label, color=df.color_label)
    assert p == "series_labelling"



import pandas as pd

from plotnine import aes, geom_polygon, ggplot

df = pd.DataFrame(
    {
        "x": ([1, 2, 3, 2] + [5, 6, 7] + [9, 9, 10, 11, 11, 10]),
        "y": ([2, 3, 2, 1] + [1, 3, 1] + [1.5, 2.5, 3, 2.5, 1.5, 1]),
        "z": ([1] * 4 + [2] * 3 + [3] * 6),
    }
)


def test_aesthetics():
    p = (
        ggplot(df, aes("x", group="factor(z)"))
        + geom_polygon(aes(y="y"))
        + geom_polygon(aes(y="y+3", alpha="z"))
        + geom_polygon(
            aes(y="y+6", linetype="factor(z)"),
            color="brown",
            fill=None,
            size=2,
        )
        + geom_polygon(aes(y="y+9", color="z"), fill=None, size=2)
        + geom_polygon(aes(y="y+12", fill="factor(z)"))
        + geom_polygon(
            aes(y="y+15", size="z"), color="yellow", show_legend=False
        )
    )

    assert p == "aesthetics"


def test_no_fill():
    p = (
        ggplot(df, aes("x", group="factor(z)"))
        + geom_polygon(aes(y="y"), fill=None, color="red", size=2)
        + geom_polygon(aes(y="y+2"), fill="None", color="green", size=2)
        + geom_polygon(aes(y="y+4"), fill="none", color="blue", size=2)
    )
    assert p == "no_fill"



from __future__ import annotations

import typing
from contextlib import suppress
from copy import copy, deepcopy
from warnings import warn

import numpy as np
import pandas as pd
from mizani.bounds import censor, expand_range_distinct, rescale, zero_range

from ..doctools import document
from ..exceptions import PlotnineError, PlotnineWarning
from ..iapi import range_view, scale_view
from ..mapping.aes import is_position_aes, rename_aesthetics
from ..utils import Registry, match
from ._expand import expand_range
from .range import Range, RangeContinuous, RangeDiscrete

if typing.TYPE_CHECKING:
    from typing import Any, Optional, Sequence, Type

    from plotnine.typing import (
        AnyArrayLike,
        CoordRange,
        FloatArrayLike,
        FloatArrayLikeTV,
        ScaleBreaks,
        ScaleBreaksRaw,
        ScaleContinuousBreaks,
        ScaleContinuousBreaksRaw,
        ScaleContinuousLimits,
        ScaleContinuousLimitsRaw,
        ScaledAestheticsName,
        ScaleDiscreteBreaks,
        ScaleDiscreteBreaksRaw,
        ScaleDiscreteLimits,
        ScaleDiscreteLimitsRaw,
        ScaleLabels,
        ScaleLabelsRaw,
        ScaleLimits,
        ScaleLimitsRaw,
        ScaleMinorBreaksRaw,
        Trans,
        TupleFloat2,
        TupleFloat4,
    )


class scale(metaclass=Registry):
    """
    Base class for all scales

    Parameters
    ----------
    breaks : array_like or callable, optional
        Major break points. Alternatively, a callable that
        takes a tuple of limits and returns a list of breaks.
        Default is to automatically calculate the breaks.
    expand : tuple, optional
        Multiplicative and additive expansion constants
        that determine how the scale is expanded. If
        specified must be of length 2 or 4. Specifically the
        values are in this order::

            (mul, add)
            (mul_low, add_low, mul_high, add_high)

        For example,

            - ``(0, 0)`` - Do not expand.
            - ``(0, 1)`` - Expand lower and upper limits by 1 unit.
            - ``(1, 0)`` - Expand lower and upper limits by 100%.
            - ``(0, 0, 0, 0)`` - Do not expand, as ``(0, 0)``.
            - ``(0, 0, 0, 1)`` - Expand upper limit by 1 unit.
            - ``(0, 1, 0.1, 0)`` - Expand lower limit by 1 unit and
              upper limit by 10%.
            - ``(0, 0, 0.1, 2)`` - Expand upper limit by 10% plus
              2 units.

        If not specified, suitable defaults are chosen.
    name : str, optional
        Name used as the label of the scale. This is what
        shows up as the axis label or legend title. Suitable
        defaults are chosen depending on the type of scale.
    labels : list or callable, optional
        List of :class:`str`. Labels at the `breaks`.
        Alternatively, a callable that takes an array_like of
        break points as input and returns a list of strings.
    limits : array_like, optional
        Limits of the scale. Most commonly, these are the
        min & max values for the scales. For scales that
        deal with categoricals, these may be a subset or
        superset of the categories.
    na_value : scalar
        What value to assign to missing values. Default
        is to assign ``np.nan``.
    palette : callable, optional
        Function to map data points onto the scale. Most
        scales define their own palettes.
    aesthetics : list, optional
        list of :class:`str`. Aesthetics covered by the
        scale. These are defined by each scale and the
        user should probably not change them. Have fun.
    """

    __base__ = True

    # aesthetics affected by this scale
    _aesthetics: list[ScaledAestheticsName] = []

    # What to do with the NA values
    na_value: Any = np.nan

    # used as the axis label or legend title
    name: str | None = None

    # major breaks
    breaks: ScaleBreaksRaw = True

    # labels at the breaks
    labels: ScaleLabelsRaw = True

    # legend or any other guide
    guide: str | None = "legend"

    # (min, max) - set by user
    _limits: ScaleLimitsRaw = None

    #: multiplicative and additive expansion constants
    expand: Optional[TupleFloat2 | TupleFloat4] = None

    # range of aesthetic, instantiated in __init__
    range: Range
    _range_class: type[Range] = Range

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)
            else:
                msg = "{} could not recognise parameter `{}`"
                warn(msg.format(self.__class__.__name__, k), PlotnineWarning)

        self.range = self._range_class()

        if np.iterable(self.breaks) and np.iterable(self.labels):
            if len(self.breaks) != len(self.labels):  # pyright: ignore
                raise PlotnineError("Breaks and labels have unequal lengths")

        if (
            self.breaks is None
            and not is_position_aes(self.aesthetics)
            and self.guide is not None
        ):
            self.guide = None

    @property
    def aesthetics(self):
        return self._aesthetics

    @aesthetics.setter
    def aesthetics(self, value):
        if isinstance(value, str):
            value = [value]
        # TODO: Find a way to make the type checking work
        self._aesthetics = rename_aesthetics(value)  # pyright: ignore

    def __radd__(self, gg):
        """
        Add this scale to ggplot object
        """
        gg.scales.append(copy(self))
        return gg

    @staticmethod
    def palette(n):
        """
        Aesthetic mapping function

        Note that not all scales need to implement/provide a palette.
        For example identity & position scales do not use a palette.
        """
        raise NotImplementedError("Not Implemented")

    def map(self, x, limits=None):
        """
        Map every element of x

        The palette should do the real work, this should
        make sure that sensible values are sent and
        return from the palette.
        """
        raise NotImplementedError("Not Implemented")

    def train(self, x):
        """
        Train scale

        Parameters
        ----------
        x: pd.series | np.array
            a column of data to train over
        """
        raise NotImplementedError("Not Implemented")

    def dimension(self, expand=None, limits=None):
        """
        Get the phyical size of the scale.
        """
        raise NotImplementedError("Not Implemented")

    def expand_limits(
        self,
        limits: ScaleLimits,
        expand: TupleFloat2 | TupleFloat4,
        coord_limits: CoordRange | None,
        trans: Trans | Type[Trans],
    ) -> range_view:
        """
        Exand the limits of the scale
        """
        raise NotImplementedError("Not Implemented")

    def transform_df(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Transform dataframe
        """
        raise NotImplementedError("Not Implemented")

    def transform(self, x):
        """
        Transform array|series x
        """
        raise NotImplementedError("Not Implemented")

    def inverse(self, x):
        """
        Inverse transform array|series x
        """
        raise NotImplementedError("Not Implemented")

    def view(
        self,
        limits: Optional[ScaleLimits] = None,
        range: Optional[CoordRange] = None,
    ) -> scale_view:
        """
        Information about the trained scale
        """
        raise NotImplementedError("Not Implemented")

    def default_expansion(
        self,
        mult: float | TupleFloat2 = 0,
        add: float | TupleFloat2 = 0,
        expand=True,
    ) -> TupleFloat4:
        """
        Get default expansion for this scale
        """
        if not expand:
            return (0, 0, 0, 0)

        if self.expand:
            if len(self.expand) == 2:
                mult, add = self.expand
            else:
                # n == 4:
                mult = self.expand[0], self.expand[2]
                add = self.expand[1], self.expand[3]

        if isinstance(mult, (float, int)):
            mult = (mult, mult)

        if isinstance(add, (float, int)):
            add = (add, add)

        if len(mult) != 2:
            raise ValueError(
                "The scale expansion multiplication factor should "
                "either be a single float, or a tuple of two floats. "
            )

        if len(add) != 2:
            raise ValueError(
                "The scale expansion addition constant should "
                "either be a single float, or a tuple of two floats. "
            )

        return (mult[0], add[0], mult[1], add[1])

    def clone(self):
        return deepcopy(self)

    def reset(self):
        """
        Set the range of the scale to None.

        i.e Forget all the training
        """
        self.range.reset()

    def is_empty(self) -> bool:
        """
        Whether the scale has size information
        """
        if not hasattr(self, "range"):
            return True
        return self.range.is_empty() and self._limits is None

    @property
    def limits(self) -> ScaleLimits:
        raise NotImplementedError("Not Implemented")

    @limits.setter
    def limits(self, value: ScaleLimitsRaw):
        raise NotImplementedError("Not Implemented")

    def train_df(self, df: pd.DataFrame):
        """
        Train scale from a dataframe
        """
        aesthetics = sorted(set(self.aesthetics) & set(df.columns))
        for ae in aesthetics:
            self.train(df[ae])

    def map_df(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Map df
        """
        if len(df) == 0:
            return df

        aesthetics = set(self.aesthetics) & set(df.columns)
        for ae in aesthetics:
            df[ae] = self.map(df[ae])

        return df

    def get_labels(self, breaks: Optional[ScaleBreaks] = None) -> ScaleLabels:
        """
        Get labels, calculating them if required
        """
        raise NotImplementedError()

    def get_breaks(self, limits: Optional[ScaleLimits] = None) -> ScaleBreaks:
        """
        Get Breaks
        """
        raise NotImplementedError()

    def get_bounded_breaks(
        self, limits: Optional[ScaleLimits] = None
    ) -> ScaleBreaks:
        """
        Return Breaks that are within the limits
        """
        raise NotImplementedError()


@document
class scale_discrete(scale):
    """
    Base class for all discrete scales

    Parameters
    ----------
    {superclass_parameters}
    limits : array_like, optional
        Limits of the scale. For scales that deal with
        categoricals, these may be a subset or superset of
        the categories. Data values that are not in the limits
        will be treated as missing data and represented with
        the ``na_value``.
    drop : bool
        Whether to drop unused categories from
        the scale
    na_translate : bool
        If ``True`` translate missing values and show them.
        If ``False`` remove missing values. Default value is
        ``True``
    na_value : object
        If ``na_translate=True``, what aesthetic value should be
        assigned to the missing values. This parameter does not
        apply to position scales where ``nan`` is always placed
        on the right.
    """

    _range_class = RangeDiscrete
    _limits: Optional[ScaleDiscreteLimitsRaw]
    range: RangeDiscrete
    breaks: ScaleDiscreteBreaksRaw
    drop: bool = True  # drop unused factor levels from the scale
    na_translate: bool = True

    @property
    def limits(self) -> ScaleDiscreteLimits:
        if self.is_empty():
            return ("0", "1")

        if self._limits is None:
            return tuple(self.range.range)
        elif callable(self._limits):
            return tuple(self._limits(self.range.range))
        else:
            return tuple(self._limits)

    @limits.setter
    def limits(self, value: ScaleDiscreteLimitsRaw):
        self._limits = value

    @staticmethod
    def palette(n: int) -> Sequence[Any]:
        """
        Aesthetic mapping function
        """
        raise NotImplementedError("Not Implemented")

    def train(self, x, drop=False):
        """
        Train scale

        Parameters
        ----------
        x: pd.series | np.array
            a column of data to train over
        drop : bool
            Whether to drop(not include) unused categories

        A discrete range is stored in a list
        """
        if not len(x):
            return

        na_rm = not self.na_translate
        self.range.train(x, drop, na_rm=na_rm)

    def dimension(self, expand=(0, 0, 0, 0), limits=None):
        """
        Get the phyical size of the scale

        Unlike limits, this always returns a numeric vector of length 2
        """
        if limits is None:
            limits = self.limits
        return expand_range_distinct(self.limits, expand)

    def expand_limits(
        self,
        limits: ScaleDiscreteLimits,
        expand: TupleFloat2 | TupleFloat4,
        coord_limits: TupleFloat2,
        trans: Trans,
    ) -> range_view:
        """
        Calculate the final range in coordinate space
        """
        # Turn discrete limits into a tuple of continuous limits
        is_empty = self.is_empty() or len(limits) == 0
        climits = (0, 1) if is_empty else (1, len(limits))
        if coord_limits is not None:
            # - Override None in coord_limits
            # - Expand limits in coordinate space
            # - Remove any computed infinite values &
            c0, c1 = coord_limits
            climits = (
                climits[0] if c0 is None else c0,
                climits[1] if c1 is None else c1,
            )
        return expand_range(climits, expand, trans)

    def view(
        self,
        limits: Optional[ScaleDiscreteLimits] = None,
        range: Optional[CoordRange] = None,
    ) -> scale_view:
        """
        Information about the trained scale
        """
        if limits is None:
            limits = self.limits

        if range is None:
            range = self.dimension(limits=limits)

        breaks_d = self.get_breaks(limits)
        breaks = self.map(pd.Categorical(breaks_d))
        minor_breaks = []
        labels = self.get_labels(breaks_d)

        sv = scale_view(
            scale=self,
            aesthetics=self.aesthetics,
            name=self.name,
            limits=limits,
            range=range,
            breaks=breaks,
            labels=labels,
            minor_breaks=minor_breaks,
        )
        return sv

    def default_expansion(self, mult=0, add=0.6, expand=True):
        """
        Get the default expansion for a discrete scale
        """
        return super().default_expansion(mult, add, expand)

    def map(
        self, x, limits: Optional[ScaleDiscreteLimits] = None
    ) -> Sequence[Any]:
        """
        Map values in x to a palette
        """
        if limits is None:
            limits = self.limits

        n = sum(~pd.isnull(list(limits)))
        pal = self.palette(n)
        if isinstance(pal, dict):
            # manual palette with specific assignments
            pal_match = []
            for val in x:
                try:
                    pal_match.append(pal[val])
                except KeyError:
                    pal_match.append(self.na_value)
        else:
            if not isinstance(pal, np.ndarray):
                pal = np.asarray(pal, dtype=object)
            idx = np.asarray(match(x, limits))
            try:
                pal_match = [pal[i] if i >= 0 else None for i in idx]
            except IndexError:
                # Deal with missing data
                # - Insert NaN where there is no match
                pal = np.hstack((pal.astype(object), np.nan))
                idx = np.clip(idx, 0, len(pal) - 1)
                pal_match = list(pal[idx])

        if self.na_translate:
            bool_pal_match = pd.isnull(pal_match)
            if len(bool_pal_match.shape) > 1:
                # linetypes take tuples, these return 2d
                bool_pal_match = bool_pal_match.any(axis=1)
            bool_idx = pd.isnull(x) | bool_pal_match
            if bool_idx.any():
                pal_match = [
                    x if i else self.na_value
                    for x, i in zip(pal_match, ~bool_idx)
                ]

        return pal_match

    def get_breaks(
        self, limits: Optional[ScaleDiscreteLimits] = None
    ) -> ScaleDiscreteBreaks:
        """
        Return an ordered list of breaks

        The form is suitable for use by the guides e.g.
            ['fair', 'good', 'very good', 'premium', 'ideal']
        """
        if limits is None:
            limits = self.limits

        if self.is_empty():
            return []

        if self.breaks is True:
            breaks = list(limits)
        elif self.breaks in (False, None):
            breaks = []
        elif callable(self.breaks):
            breaks = self.breaks(limits)
        else:
            _wanted_breaks = set(self.breaks)
            breaks = [l for l in limits if l in _wanted_breaks]

        return breaks

    def get_bounded_breaks(
        self, limits: Optional[ScaleDiscreteLimits] = None
    ) -> ScaleDiscreteBreaks:
        """
        Return Breaks that are within limits
        """
        if limits is None:
            limits = self.limits

        lookup = set(limits)
        breaks = self.get_breaks()
        strict_breaks = [b for b in breaks if b in lookup]
        return strict_breaks

    def get_labels(
        self, breaks: Optional[ScaleDiscreteBreaks] = None
    ) -> ScaleLabels:
        """
        Generate labels for the legend/guide breaks
        """
        if self.is_empty():
            return []

        if breaks is None:
            breaks = self.get_breaks()

        # The labels depend on the breaks if the breaks.
        # No breaks, no labels
        if breaks in (None, False) or self.labels in (None, False):
            return []
        elif self.labels is True:
            return [str(b) for b in breaks]
        elif callable(self.labels):
            return self.labels(breaks)
        # if a dict is used to rename some labels
        elif isinstance(self.labels, dict):
            labels = [
                str(self.labels[b]) if b in self.labels else str(b)
                for b in breaks
            ]
            return labels
        else:
            return self.labels

    def transform_df(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Transform dataframe
        """
        # Discrete scales do not do transformations
        return df

    def transform(self, x):
        """
        Transform array|series x
        """
        # Discrete scales do not do transformations
        return x


@document
class scale_continuous(scale):
    """
    Base class for all continuous scales

    Parameters
    ----------
    {superclass_parameters}
    trans : str | function
        Name of a trans function or a trans function.
        See :mod:`mizani.transforms` for possible options.
    oob : function
        Function to deal with out of bounds (limits)
        data points. Default is to turn them into
        ``np.nan``, which then get dropped.
    minor_breaks : list-like or int or callable or None
        If a list-like, it is the minor breaks points.
        If an integer, it is the number of minor breaks between
        any set of major breaks.
        If a function, it should have the signature
        ``func(limits)`` and return a list-like of consisting
        of the minor break points.
        If ``None``, no minor breaks are calculated.
        The default is to automatically calculate them.
    rescaler : function, optional
        Function to rescale data points so that they can
        be handled by the palette. Default is to rescale
        them onto the [0, 1] range. Scales that inherit
        from this class may have another default.

    Notes
    -----
    If using the class directly all arguments must be
    keyword arguments.
    """

    _range_class = RangeContinuous
    _limits: Optional[ScaleContinuousLimitsRaw]
    rescaler = staticmethod(rescale)  # Used by diverging & n colour gradients
    oob = staticmethod(censor)  # what to do with out of bounds data points
    breaks: ScaleContinuousBreaksRaw
    minor_breaks: ScaleMinorBreaksRaw = True
    _trans: Trans | str = "identity"  # transform class

    def __init__(self, **kwargs):
        # Make sure we have a transform.
        self.trans = kwargs.pop("trans", self._trans)

        with suppress(KeyError):
            self.limits = kwargs.pop("limits")

        scale.__init__(self, **kwargs)

    def _check_trans(self, t):
        """
        Check if transform t is valid

        When scales specialise on a specific transform (other than
        the identity transform), the user should know when they
        try to change the transform.

        Parameters
        ----------
        t : mizani.transforms.trans
            Transform object
        """
        orig_trans_name = self.__class__._trans
        new_trans_name = t.__class__.__name__
        if new_trans_name.endswith("_trans"):
            new_trans_name = new_trans_name[:-6]
        if orig_trans_name != "identity":
            if new_trans_name != orig_trans_name:
                warn(
                    "You have changed the transform of a specialised scale. "
                    "The result may not be what you expect.\n"
                    "Original transform: {}\n"
                    "New transform: {}".format(
                        orig_trans_name, new_trans_name
                    ),
                    PlotnineWarning,
                )

    @property
    def trans(self) -> Trans:
        return self._trans  # pyright: ignore

    @trans.setter
    def trans(self, value: Trans | str | Type[Trans]):
        from mizani.transforms import gettrans

        t: Trans = gettrans(value)
        self._check_trans(t)
        self._trans = t
        self._trans.aesthetic = self.aesthetics[0]  # pyright: ignore

    @property
    def limits(self):
        if self.is_empty():
            return (0, 1)

        if self._limits is None:
            return self.range.range
        elif callable(self._limits):
            # Function works in the dataspace, but the limits are
            # stored in transformed space. The range of the scale is
            # in transformed space (i.e. with in the domain of the scale)
            _range = self.trans.inverse(self.range.range)
            return tuple(self.trans.transform(self._limits(_range)))
        elif self._limits is not None and not self.range.is_empty():
            # Fall back to the range if the limits
            # are not set or if any is None or NaN
            if len(self._limits) == len(self.range.range):
                return tuple(
                    self.trans.transform(r) if pd.isnull(l) else l
                    for l, r in zip(self._limits, self.range.range)
                )
        return tuple(self._limits)

    @limits.setter
    def limits(self, value: ScaleContinuousLimitsRaw):
        """
        Limits for the continuous scale

        Parameters
        ----------
        value : array-like | callable
            Limits in the dataspace.
        """
        # Notes
        # -----
        # The limits are given in original dataspace
        # but they are stored in transformed space since
        # all computations happen on transformed data. The
        # labeling of the plot axis and the guides are in
        # the original dataspace.
        if isinstance(value, bool) or value is None or callable(value):
            self._limits = value
            return

        a, b = value
        a, b = (
            self.trans.transform(a) if a is not None else a,
            self.trans.transform(b) if b is not None else b,
        )
        with suppress(TypeError):
            if a > b:
                a, b = b, a

        self._limits = a, b

    def train(self, x):
        """
        Train scale

        Parameters
        ----------
        x: pd.series | np.array
            a column of data to train over
        """
        if not len(x):
            return

        self.range.train(x)

    def transform_df(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Transform dataframe
        """
        if len(df) == 0:
            return df

        aesthetics = set(self.aesthetics) & set(df.columns)
        for ae in aesthetics:
            with suppress(TypeError):
                df[ae] = self.transform(df[ae])

        return df

    def transform(self, x: FloatArrayLikeTV) -> FloatArrayLikeTV:
        """
        Transform array|series x
        """
        try:
            return self.trans.transform(x)
        except TypeError:
            return [self.trans.transform(val) for val in x]  # pyright: ignore

    def inverse(self, x: FloatArrayLikeTV) -> FloatArrayLikeTV:
        """
        Inverse transform array|series x
        """
        try:
            return self.trans.inverse(x)
        except TypeError:
            return [self.trans.inverse(val) for val in x]  # pyright: ignore

    def dimension(self, expand=(0, 0, 0, 0), limits=None):
        """
        Get the phyical size of the scale

        Unlike limits, this always returns a numeric vector of length 2
        """
        if limits is None:
            limits = self.limits
        return expand_range_distinct(limits, expand)

    def expand_limits(
        self,
        limits: ScaleContinuousLimits,
        expand: TupleFloat2 | TupleFloat4,
        coord_limits: CoordRange | None,
        trans: Trans,
    ) -> range_view:
        """
        Calculate the final range in coordinate space
        """
        # - Override None in coord_limits
        # - Expand limits in coordinate space
        # - Remove any computed infinite values &
        if coord_limits is not None:
            c0, c1 = coord_limits
            limits = (
                limits[0] if c0 is None else c0,
                limits[1] if c1 is None else c1,
            )
        return expand_range(limits, expand, trans)

    def view(
        self,
        limits: Optional[CoordRange] = None,
        range: Optional[CoordRange] = None,
    ) -> scale_view:
        """
        Information about the trained scale
        """
        if limits is None:
            limits = self.limits

        if range is None:
            range = self.dimension(limits=limits)

        breaks = self.get_bounded_breaks(range)
        labels = self.get_labels(breaks)

        ubreaks = self.get_breaks(range)
        minor_breaks = self.get_minor_breaks(ubreaks, range)

        sv = scale_view(
            scale=self,
            aesthetics=self.aesthetics,
            name=self.name,
            limits=limits,
            range=range,
            breaks=breaks,
            labels=labels,
            minor_breaks=minor_breaks,
        )
        return sv

    def default_expansion(self, mult=0.05, add=0, expand=True):
        """
        Get the default expansion for continuous scale
        """
        return super().default_expansion(mult, add, expand)

    @staticmethod
    def palette(arr: FloatArrayLike) -> Sequence[Any]:
        """
        Aesthetic mapping function
        """
        raise NotImplementedError("Not Implemented")

    def map(
        self, x: AnyArrayLike, limits: Optional[ScaleContinuousLimits] = None
    ) -> AnyArrayLike:
        if limits is None:
            limits = self.limits

        x = self.oob(self.rescaler(x, _from=limits))

        uniq = np.unique(x)
        pal = np.asarray(self.palette(uniq))
        scaled = pal[match(x, uniq)]
        if scaled.dtype.kind == "U":
            scaled = [self.na_value if x == "nan" else x for x in scaled]
        else:
            scaled[pd.isnull(scaled)] = self.na_value
        return scaled

    def get_breaks(
        self, limits: Optional[ScaleContinuousLimits] = None
    ) -> ScaleContinuousBreaks:
        """
        Generate breaks for the axis or legend

        Parameters
        ----------
        limits : list-like | None
            If None the self.limits are used
            They are expected to be in transformed
            space.

        Returns
        -------
        out : array-like

        Notes
        -----
        Breaks are calculated in data space and
        returned in transformed space since all
        data is plotted in transformed space.
        """
        if limits is None:
            limits = self.limits

        # To data space
        _limits = self.inverse(limits)

        if self.is_empty():
            breaks = []
        elif self.breaks is True:
            # TODO: Fix this type mismatch in mizani with
            # a typevar so that type-in = type-out
            _tlimits = self.trans.breaks(_limits)
            breaks: ScaleContinuousBreaks = _tlimits  # pyright: ignore
        elif self.breaks is False or self.breaks is None:
            breaks = []
        elif zero_range(_limits):
            breaks = [_limits[0]]
        elif callable(self.breaks):
            breaks = self.breaks(_limits)
        else:
            breaks = self.breaks

        breaks = self.transform(breaks)
        return breaks

    def get_bounded_breaks(
        self, limits: Optional[ScaleContinuousLimits] = None
    ) -> ScaleContinuousBreaks:
        """
        Return Breaks that are within limits
        """
        if limits is None:
            limits = self.limits
        breaks = self.get_breaks(limits)
        strict_breaks = [b for b in breaks if limits[0] <= b <= limits[1]]
        return strict_breaks

    def get_minor_breaks(
        self,
        major: ScaleContinuousBreaks,
        limits: Optional[ScaleContinuousLimits] = None,
    ) -> ScaleContinuousBreaks:
        """
        Return minor breaks
        """
        if limits is None:
            limits = self.limits

        if self.minor_breaks is True:
            # TODO: Remove ignore when mizani is static typed
            minor_breaks: ScaleContinuousBreaks = self.trans.minor_breaks(
                major, limits
            )  # pyright: ignore
        elif isinstance(self.minor_breaks, int):
            # TODO: Remove ignore when mizani is static typed
            minor_breaks: ScaleContinuousBreaks = self.trans.minor_breaks(
                major, limits, n=self.minor_breaks
            )  # pyright: ignore
        elif self.minor_breaks in (False, None) or not len(major):
            minor_breaks = []
        elif callable(self.minor_breaks):
            breaks = self.minor_breaks(self.trans.inverse(limits))
            _major = set(major)
            minor = self.trans.transform(breaks)
            minor_breaks = [x for x in minor if x not in _major]
        else:
            minor_breaks = self.minor_breaks

        return minor_breaks

    def get_labels(
        self, breaks: Optional[ScaleContinuousBreaks] = None
    ) -> ScaleLabels:
        """
        Generate labels for the axis or legend

        Parameters
        ----------
        breaks: None or array-like
            If None, use self.breaks.
        """
        if breaks is None:
            breaks = self.get_breaks()

        breaks = self.inverse(breaks)

        if self.labels is True:
            labels: Sequence[str] = self.trans.format(breaks)
        elif self.labels in (False, None):
            labels = []
        elif callable(self.labels):
            labels = self.labels(breaks)
        elif isinstance(self.labels, dict):
            labels = [
                str(self.labels[b]) if b in self.labels else str(b)
                for b in breaks
            ]
        else:
            # When user sets breaks and labels of equal size,
            # but the limits exclude some of the breaks.
            # We remove the corresponding labels
            from collections.abc import Sized

            labels = self.labels
            if (
                len(labels) != len(breaks)
                and isinstance(self.breaks, Sized)
                and len(labels) == len(self.breaks)
            ):
                _wanted_breaks = set(breaks)
                labels = [
                    l
                    for l, b in zip(labels, self.breaks)
                    if b in _wanted_breaks
                ]

        if len(labels) != len(breaks):
            raise PlotnineError("Breaks and labels are different lengths")

        return labels


@document
class scale_datetime(scale_continuous):
    """
    Base class for all date/datetime scales

    Parameters
    ----------
    date_breaks : str
        A string giving the distance between major breaks.
        For example `'2 weeks'`, `'5 years'`. If specified,
        ``date_breaks`` takes precedence over
        ``breaks``.
    date_labels : str
        Format string for the labels.
        See :ref:`strftime <strftime-strptime-behavior>`.
        If specified, ``date_labels`` takes precedence over
        ``labels``.
    date_minor_breaks : str
        A string giving the distance between minor breaks.
        For example `'2 weeks'`, `'5 years'`. If specified,
        ``date_minor_breaks`` takes precedence over
        ``minor_breaks``.
    {superclass_parameters}
    """

    _trans = "datetime"

    def __init__(self, **kwargs):
        from mizani.breaks import date_breaks
        from mizani.formatters import date_format

        # Permit the use of the general parameters for
        # specifying the format strings
        with suppress(KeyError):
            breaks = kwargs["breaks"]
            if isinstance(breaks, str):
                kwargs["breaks"] = date_breaks(breaks)

        with suppress(KeyError):
            minor_breaks = kwargs["minor_breaks"]
            if isinstance(minor_breaks, str):
                kwargs["minor_breaks"] = date_breaks(minor_breaks)

        # Using the more specific parameters take precedence
        with suppress(KeyError):
            breaks_fmt = kwargs.pop("date_breaks")
            kwargs["breaks"] = date_breaks(breaks_fmt)

        with suppress(KeyError):
            labels_fmt = kwargs.pop("date_labels")
            kwargs["labels"] = date_format(labels_fmt)

        with suppress(KeyError):
            minor_breaks_fmt = kwargs.pop("date_minor_breaks")
            kwargs["minor_breaks"] = date_breaks(minor_breaks_fmt)

        scale_continuous.__init__(self, **kwargs)
